import json
from unittest import mock
from unittest.mock import AsyncMock

from fastapi import HTTPException, WebSocketDisconnect, status
from pytest import raises

from orchestrator.websocket import broadcast_invalidate_cache, sync_broadcast_invalidate_cache


def test_websocket_events_ping_pong(test_client):
    with test_client.websocket_connect("api/ws/events") as websocket:
        websocket.send_text("__ping__")
        assert websocket.receive_text() == "__pong__"


def test_websocket_events_invalidate_cache(test_client):
    with test_client.websocket_connect("api/ws/events") as websocket:
        sync_broadcast_invalidate_cache({"type": "foobar"})
        assert websocket.receive_json() == {"name": "invalidateCache", "value": {"type": "foobar"}}


async def test_websocket_events_invalidate_cache_async(test_client):
    with test_client.websocket_connect("api/ws/events") as websocket:
        await broadcast_invalidate_cache({"type": "foobar", "id": "bar"})
        assert websocket.receive_json() == {"name": "invalidateCache", "value": {"type": "foobar", "id": "bar"}}


@mock.patch("orchestrator.websocket.websocket_manager.authorize_websocket")
@mock.patch("orchestrator.websocket.websocket_manager.authenticate_websocket")
async def test_websocket_events_invalid_protocol(mock_user, mock_security, test_client):
    # when: we create a websocket connection with an invalid protocol
    with raises(WebSocketDisconnect) as error_info:
        with test_client.websocket_connect(
            "api/ws/events", headers={"sec-websocket-protocol": "my token"}
        ) as websocket:
            websocket.send_text("__ping__")
            websocket.receive_text()

    # then: it returns an error with code 1002
    assert error_info.value.code == status.WS_1002_PROTOCOL_ERROR


@mock.patch("orchestrator.websocket.websocket_manager.authorize_websocket")
@mock.patch("orchestrator.websocket.websocket_manager.authenticate_websocket")
async def test_websocket_events_not_authenticated(mock_authenticate_websocket, mock_authorize_websocket, test_client):
    # given: the user is not authenticated
    mock_authenticate_websocket.side_effect = AsyncMock(side_effect=HTTPException(status_code=401))
    mock_authorize_websocket.side_effect = AsyncMock(side_effect=Exception("This should not be called"))

    # when: we create a websocket connection and ping it
    with test_client.websocket_connect("api/ws/events") as websocket:
        websocket.send_text("__ping__")
        reply = websocket.receive_text()

    # then: it does not reply with pong and returns 401
    assert reply != "__pong__"
    assert json.loads(reply) == {"error": {"detail": "Unauthorized", "headers": None, "status_code": 401}}
    assert mock_authenticate_websocket.call_count == 1
    assert mock_authenticate_websocket.call_args[1]["token"] == ""
    assert mock_authorize_websocket.call_count == 0


@mock.patch("orchestrator.websocket.websocket_manager.authorize_websocket")
@mock.patch("orchestrator.websocket.websocket_manager.authenticate_websocket")
async def test_websocket_events_not_authorized(mock_user, mock_security, test_client):
    # given: the user is authenticated but not authorized
    mock_user.side_effect = AsyncMock(return_value={"active": True})
    mock_security.side_effect = AsyncMock(side_effect=HTTPException(status_code=403))

    # when: we create a websocket connection and ping it
    with test_client.websocket_connect(
        "api/ws/events", headers={"sec-websocket-protocol": "base64.bearer.token, my token"}
    ) as websocket:
        websocket.send_text("__ping__")
        reply = websocket.receive_text()

    # then: it does not reply with pong and returns 403
    assert reply != "__pong__"
    assert json.loads(reply) == {"error": {"detail": "Forbidden", "headers": None, "status_code": 403}}
    assert mock_user.call_count == 1
    assert mock_user.call_args[1]["token"] == "my token"  # noqa: S105
    assert mock_security.call_count == 1


@mock.patch("orchestrator.websocket.websocket_manager.authorize_websocket")
@mock.patch("orchestrator.websocket.websocket_manager.authenticate_websocket")
async def test_websocket_events_authorized(mock_user, mock_security, test_client):
    # given: the user is authenticated and authorized
    mock_user.side_effect = AsyncMock(return_value={"active": True})
    mock_security.side_effect = AsyncMock(return_value=True)

    # when: we create a websocket connection and ping it
    with test_client.websocket_connect(
        "api/ws/events", headers={"sec-websocket-protocol": "base64.bearer.token, my token"}
    ) as websocket:
        websocket.send_text("__ping__")
        reply = websocket.receive_text()

    # then: it replies with pong
    assert reply == "__pong__"
    assert mock_user.call_args[1]["token"] == "my token"  # noqa: S105
    assert mock_security.call_count == 1



import structlog
from orchestrator.domain import SubscriptionModel
from orchestrator.forms import FormPage
from orchestrator.forms.validators import CustomerId, Divider, Label
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle, UUIDstr
from orchestrator.workflow import StepList, begin, step
from orchestrator.workflows.steps import store_process_subscription
from orchestrator.workflows.utils import create_workflow
from pydantic import ConfigDict

from products.product_types.example4 import Example4Inactive, Example4Provisioning


def subscription_description(subscription: SubscriptionModel) -> str:
    """Generate subscription description.

    The suggested pattern is to implement a subscription service that generates a subscription specific
    description, in case that is not present the description will just be set to the product name.
    """
    return f"{subscription.product.name} subscription"


logger = structlog.get_logger(__name__)


def initial_input_form_generator(product_name: str) -> FormGenerator:
    # TODO add additional fields to form if needed

    class CreateExample4Form(FormPage):
        model_config = ConfigDict(title=product_name)

        customer_id: CustomerId

        example4_settings: Label
        divider_1: Divider

        num_val: int | None = None

    user_input = yield CreateExample4Form
    user_input_dict = user_input.dict()

    return user_input_dict


@step("Construct Subscription model")
def construct_example4_model(
    product: UUIDstr,
    customer_id: UUIDstr,
    num_val: int | None,
) -> State:
    example4 = Example4Inactive.from_product_id(
        product_id=product,
        customer_id=customer_id,
        status=SubscriptionLifecycle.INITIAL,
    )
    example4.example4.num_val = num_val

    example4 = Example4Provisioning.from_other_lifecycle(example4, SubscriptionLifecycle.PROVISIONING)
    example4.description = subscription_description(example4)

    return {
        "subscription": example4,
        "subscription_id": example4.subscription_id,  # necessary to be able to use older generic step functions
        "subscription_description": example4.description,
    }


additional_steps = begin


@create_workflow("Create example4", initial_input_form=initial_input_form_generator, additional_steps=additional_steps)
def create_example4() -> StepList:
    return (
        begin >> construct_example4_model >> store_process_subscription(Target.CREATE)
        # TODO add provision step(s)
    )



import json
from hashlib import md5
from http import HTTPStatus

from redis import Redis

from orchestrator import app_settings
from orchestrator.utils.redis import ONE_WEEK
from test.unit_tests.config import GRAPHQL_ENDPOINT, GRAPHQL_HEADERS


def get_settings_query() -> bytes:
    query = """
query SettingsQuery {
  settings {
    cacheNames
    engineSettings {
      globalLock
      globalStatus
      runningProcesses
    }
    workerStatus {
      executorType
      numberOfQueuedJobs
      numberOfWorkersOnline
      numberOfRunningJobs
    }
  }
}
    """
    return bytes(json.dumps({"operationName": "SettingsQuery", "query": query}), "utf-8")


def get_clear_cache_mutation(name: str = "all"):
    query = """
mutation ClearCacheMutation($name: String!) {
  clearCache(name: $name) {
    __typename
    ... on CacheClearSuccess {
      deleted
    }
    ... on Error {
      message
    }
  }
}
    """
    return json.dumps({"operationName": "ClearCacheMutation", "query": query, "variables": {"name": name}}).encode(
        "utf-8"
    )


def get_test_set_global_lock_mutation(global_lock: bool = False):
    query = """
mutation UpdateStatusMutation($globalLock: Boolean = false) {
  updateStatus(globalLock: $globalLock) {
    __typename
    ... on EngineSettingsType {
      globalStatus
      globalLock
      runningProcesses
    }
    ... on Error {
      message
    }
  }
}
    """
    return json.dumps(
        {"operationName": "UpdateStatusMutation", "query": query, "variables": {"globalLock": global_lock}}
    ).encode("utf-8")


def test_settings_query(test_client):
    response = test_client.post(
        GRAPHQL_ENDPOINT,
        content=get_settings_query(),
        headers=GRAPHQL_HEADERS,
    )
    assert HTTPStatus.OK == response.status_code
    result = response.json()
    settings_data = result["data"]["settings"]
    worker_status = settings_data["workerStatus"]

    assert settings_data["cacheNames"] == {"all": "All caches"}
    assert settings_data["engineSettings"] == {"globalLock": False, "globalStatus": "RUNNING", "runningProcesses": 0}
    assert worker_status["executorType"] == "threadpool"
    assert worker_status["numberOfQueuedJobs"] == 0
    assert worker_status["numberOfWorkersOnline"] == 5
    # numberOfRunningJobs is different depending if you run the whole suite vs just the graphql tests


def test_clear_cache_mutation_fails_auth(test_client, monkeypatch):
    from oauth2_lib.settings import oauth2lib_settings

    monkeypatch.setattr(oauth2lib_settings, "ENVIRONMENT_IGNORE_MUTATION_DISABLED", [])

    # oauth2lib_settings.ENVIRONMENT_IGNORE_MUTATION_DISABLED = []
    data = get_clear_cache_mutation()
    response = test_client.post("/api/graphql", content=data, headers=GRAPHQL_HEADERS)
    assert response.status_code == HTTPStatus.OK
    result = response.json()

    assert result["data"] is None
    assert result["errors"][0]["message"] == "User is not authenticated"


def test_success_clear_cache(test_client, cache_fixture):
    cache = Redis.from_url(str(app_settings.CACHE_URI))
    key = "some_model_uuid"
    test_data = {key: {"data": [1, 2, 3]}}

    # Add test data to cache
    cache.set(f"orchestrator:{key}", json.dumps(test_data), ex=ONE_WEEK)
    cache.set(
        f"orchestrator:etag:{key}",
        md5(json.dumps(test_data).encode("utf-8"), usedforsecurity=False).hexdigest(),  # noqa: S303
        ex=ONE_WEEK,
    )
    cache_fixture.extend([f"orchestrator:{key}", f"orchestrator:etag:{key}"])

    assert len(cache.keys()) == 2

    response = test_client.post("/api/graphql", content=get_clear_cache_mutation(), headers=GRAPHQL_HEADERS)
    assert HTTPStatus.OK == response.status_code

    result = response.json()
    data = result["data"]["clearCache"]
    assert data["__typename"] == "CacheClearSuccess"
    assert len(cache.keys()) == 0
    assert data["deleted"] == 2


def test_failure_clear_cache(test_client):
    response = test_client.post(GRAPHQL_ENDPOINT, content=get_clear_cache_mutation("invalid"), headers=GRAPHQL_HEADERS)
    assert HTTPStatus.OK == response.status_code
    result = response.json()

    assert result["data"] == {"clearCache": {"__typename": "Error", "message": "Invalid cache name"}}


def test_mutate_engine_global_settings(test_client):
    # Turn it off
    response = test_client.post(
        GRAPHQL_ENDPOINT, content=get_test_set_global_lock_mutation(True), headers=GRAPHQL_HEADERS
    )
    assert HTTPStatus.OK == response.status_code

    result = response.json()

    status_data = result["data"]["updateStatus"]
    assert status_data["__typename"] == "EngineSettingsType"
    assert status_data["globalStatus"] == "PAUSED"
    assert status_data["globalLock"] is True

    # Turn it back on
    response = test_client.post(
        GRAPHQL_ENDPOINT, content=get_test_set_global_lock_mutation(False), headers=GRAPHQL_HEADERS
    )
    assert HTTPStatus.OK == response.status_code

    result = response.json()

    status_data = result["data"]["updateStatus"]
    assert status_data["__typename"] == "EngineSettingsType"
    assert status_data["globalStatus"] == "RUNNING"
    assert status_data["globalLock"] is False



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ipaddress import IPv4Address, IPv6Address

import structlog

from orchestrator.devtools.populator import Populator
from orchestrator.types import UUIDstr

logger = structlog.get_logger(__name__)


class TestProductPopulator(Populator):
    def __init__(
        self,
        an_int: int,
        a_str: str,
        a_bool: bool,
        an_uuid: UUIDstr,
        an_ipv4: IPv4Address,
        an_ipv6: IPv6Address,
    ):
        self.log = logger.bind()

        self.product_name = "Test Product"
        self.endpoint_description = "Test Product Endpoint"
        self.log = logger.bind(customer="Test")

        self.an_int = an_int
        self.a_str = a_str
        self.a_bool = a_bool
        self.an_uuid = an_uuid
        self.an_ipv4 = str(an_ipv4)
        self.an_ipv6 = str(an_ipv6)

        super().__init__(self.product_name)

    def add_default_values(self) -> None:
        super().add_default_values()
        self.default_input_values.update(
            {
                "an_int": self.an_int,  # type: ignore
                "a_str": self.a_str,
                "a_bool": self.a_bool,  # type: ignore
                "an_uuid": self.an_uuid,
                "an_ipv4": self.an_ipv4,
                "an_ipv6": self.an_ipv6,
            }
        )



# Copyright 2022 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import json
from http import HTTPStatus

import pytest
from sqlalchemy import select

from orchestrator.db import SubscriptionMetadataTable, db
from orchestrator.db.models import SubscriptionCustomerDescriptionTable, SubscriptionTable
from orchestrator.domain.base import SubscriptionModel
from orchestrator.types import SubscriptionLifecycle
from test.unit_tests.conftest import do_refresh_subscriptions_search_view

subscription_fields = [
    "endDate",
    "description",
    "subscriptionId",
    "startDate",
    "status",
    "insync",
    "note",
    "product",
    "productBlockInstances",
]
subscription_product_fields = [
    "productId",
    "name",
    "description",
    "productType",
    "status",
    "tag",
    "createdAt",
    "endDate",
]
CUSTOMER_ID = "a6274018-0580-41cd-9fe8-17b946706e9f"


def build_subscriptions_query(body: str) -> str:
    return f"""
query SubscriptionQuery(
    $first: Int!,
    $after: Int!,
    $sortBy: [GraphqlSort!],
    $filterBy: [GraphqlFilter!],
    $query: String
) {{
    subscriptions(first: $first, after: $after, sortBy: $sortBy, filterBy: $filterBy, query: $query)
    {body}
}}
"""


def get_subscriptions_query(
    first: int = 10,
    after: int = 0,
    filter_by: list[dict] | None = None,
    sort_by: list | None = None,
    query_string: str | None = None,
) -> bytes:
    gql_query = build_subscriptions_query(
        """{
    page {
      description
      subscriptionId
      status
      insync
      note
      startDate
      endDate
      productBlockInstances {
        id
        parent
        subscriptionInstanceId
        ownerSubscriptionId
        productBlockInstanceValues
        inUseByRelations
      }
      product {
        productId
        name
        description
        productType
        status
        tag
        createdAt
        endDate
      }
      customerDescriptions {
        subscriptionId
        customerId
        description
      }
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
"""
    )
    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": gql_query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
            },
        }
    ).encode("utf-8")


def get_subscriptions_query_with_relations(
    first: int = 10,
    after: int = 0,
    filter_by: list[dict[str, str]] | None = None,
    sort_by: list[dict[str, str]] | None = None,
    query_string: str | None = None,
) -> bytes:
    query = build_subscriptions_query(
        """{
    page {
      description
      subscriptionId
      status
      insync
      note
      startDate
      endDate
      productBlockInstances {
        ownerSubscriptionId
        inUseByRelations
      }
      processes(sortBy: [{field: "startedAt", order: ASC}]) {
        page {
          processId
          isTask
          lastStep
          lastStatus
          assignee
          failedReason
          traceback
          workflowName
          createdBy
          startedAt
          lastModifiedAt
          product {
            productId
            name
            description
            productType
            status
            tag
            createdAt
            endDate
          }
        }
      }
      dependsOnSubscriptions {
        page {
          description
          subscriptionId
          status
          insync
          note
          startDate
          endDate
        }
      }
      inUseBySubscriptions {
        page {
          description
          subscriptionId
          status
          insync
          note
          startDate
          endDate
        }
      }
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
    """
    )
    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
            },
        }
    ).encode("utf-8")


def get_subscriptions_product_block_json_schema_query(
    first: int = 10,
    after: int = 0,
    filter_by: list[dict[str, str]] | None = None,
    sort_by: list[dict[str, str]] | None = None,
    query_string: str | None = None,
) -> bytes:
    query = build_subscriptions_query(
        """{
    page {
      _schema
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
    """
    )
    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
            },
        }
    ).encode("utf-8")


def get_subscriptions_product_generic_one(
    first: int = 10,
    after: int = 0,
    filter_by: list[str] | None = None,
    sort_by: list[dict[str, str]] | None = None,
    query_string: str | None = None,
) -> bytes:
    query = build_subscriptions_query(
        """{
    page {
      ... on GenericProductOneSubscription {
        description
        subscriptionId
        status
        insync
        note
        startDate
        endDate
        pb1 {
          rt1
        }
        pb2 {
          rt2
          rt3
        }
        customerDescriptions {
            subscriptionId
            customerId
            description
        }
      }
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
    """
    )
    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
            },
        }
    ).encode("utf-8")


def get_subscriptions_product_sub_list_union(
    first: int = 10,
    after: int = 0,
    filter_by: list[dict[str, str]] | None = None,
    sort_by: list[str] | None = None,
    query_string: str | None = None,
) -> bytes:
    query = build_subscriptions_query(
        """{
    page {
      ... on ProductSubListUnionSubscription {
        description
        subscriptionId
        status
        insync
        note
        startDate
        endDate
        product{
          productType
        }
        testBlock {
          intField
          strField
          listField
          listUnionBlocks {
            ... on SubBlockOneForTestInactive {
              intField
              strField
            }
            ... on SubBlockTwoForTestInactive {
              intField2
            }
          }
        }
      }
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
    """
    )

    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
            },
        }
    ).encode("utf-8")


def get_subscriptions_with_metadata_and_schema_query(
    first: int = 1,
    after: int = 0,
    filter_by: list[dict[str, str]] | None = None,
    sort_by: list[dict[str, str]] | None = None,
    query_string: str | None = None,
) -> bytes:
    gql_query = build_subscriptions_query(
        """{
    page {
      subscriptionId
      metadata
      _metadataSchema
    }
  }
    """
    )
    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": gql_query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
            },
        }
    ).encode("utf-8")


def test_subscriptions_single_page(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(4)
    data = get_subscriptions_query()
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    assert "errors" not in result

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 6,
        "totalItems": 7,
    }

    for subscription in subscriptions:
        for field in subscription_fields:
            assert field in subscription
        for field in subscription_product_fields:
            assert field in subscription["product"]


def test_subscriptions_has_next_page(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)
    data = get_subscriptions_query()
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    assert "errors" not in result

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": 33,
    }

    for subscription in subscriptions:
        for field in subscription_fields:
            assert field in subscription


def test_subscriptions_has_previous_page(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)
    data = get_subscriptions_query(after=1)
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    assert "errors" not in result

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert len(subscriptions) == 10

    assert pageinfo == {
        "hasPreviousPage": True,
        "hasNextPage": True,
        "startCursor": 1,
        "endCursor": 10,
        "totalItems": 33,
    }


def test_subscriptions_sorting_asc(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)
    data = get_subscriptions_query(sort_by=[{"field": "startDate", "order": "ASC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    assert "errors" not in result

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": 33,
    }

    for i in range(0, 8):
        assert subscriptions[i]["startDate"] < subscriptions[i + 1]["startDate"]


def test_subscriptions_sorting_desc(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)
    data = get_subscriptions_query(sort_by=[{"field": "startDate", "order": "DESC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    assert "errors" not in result

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": 33,
    }

    for i in range(0, 8):
        assert subscriptions[i + 1]["startDate"] < subscriptions[i]["startDate"]


def test_subscriptions_sorting_product_tag_asc(test_client, generic_subscription_1, generic_subscription_2):
    # when

    data = get_subscriptions_query(sort_by=[{"field": "tag", "order": "ASC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    assert "errors" not in result

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 4,
        "totalItems": 5,
    }

    product_tag_list = [subscription["product"]["tag"] for subscription in subscriptions]
    assert product_tag_list == ["GEN1", "GEN2", "Sub", "Sub", "UnionSub"]


def test_subscriptions_sorting_product_tag_desc(test_client, generic_subscription_1, generic_subscription_2):
    # when

    data = get_subscriptions_query(sort_by=[{"field": "tag", "order": "DESC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    assert "errors" not in result

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 4,
        "totalItems": 5,
    }

    product_tag_list = [subscription["product"]["tag"] for subscription in subscriptions]
    assert product_tag_list == ["UnionSub", "Sub", "Sub", "GEN2", "GEN1"]


def test_subscriptions_sorting_invalid_field(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)
    data = get_subscriptions_query(sort_by=[{"field": "start_date", "order": "DESC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    pageinfo = subscriptions_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": 33,
    }

    assert "errors" in result
    assert result["errors"] == [
        {
            "message": (
                "Invalid sort arguments (invalid_sorting=['start_date'] "
                "valid_sort_keys=['customerId', 'description', 'endDate', "
                "'insync', 'note', 'productCreatedAt', 'productDescription', "
                "'productEndDate', 'productId', 'productName', 'productStatus', "
                "'productTag', 'productType', 'startDate', 'status', "
                "'subscriptionId', 'tag'])"
            ),
            "path": ["subscriptions"],
            "extensions": {"error_type": "internal_error"},
        }
    ]


def test_subscriptions_sorting_invalid_order(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)
    data = get_subscriptions_query(sort_by=[{"field": "start_date", "order": "test"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()

    assert not result["data"]
    assert "errors" in result
    assert "Value 'test' does not exist in 'SortOrder'" in result["errors"][0]["message"]


@pytest.mark.parametrize(
    "query_args",
    [
        {"filter_by": [{"field": "status", "value": SubscriptionLifecycle.TERMINATED}]},
        {"query_string": "status:terminated"},
    ],
)
def test_subscriptions_filtering_on_status(
    test_client, product_type_1_subscriptions_factory, generic_product_type_1, query_args
):
    # when

    subscription_ids = product_type_1_subscriptions_factory(30)

    GenericProductOneInactive, GenericProductOne = generic_product_type_1
    subscription_1 = GenericProductOne.from_subscription(subscription_ids[0])
    subscription_1 = subscription_1.from_other_lifecycle(subscription_1, SubscriptionLifecycle.TERMINATED)
    subscription_1.save()
    subscription_9 = GenericProductOne.from_subscription(subscription_ids[8])
    subscription_9 = subscription_9.from_other_lifecycle(subscription_9, SubscriptionLifecycle.TERMINATED)
    subscription_9.save()
    do_refresh_subscriptions_search_view()

    subscription_query = get_subscriptions_query(**query_args)

    response = test_client.post(
        "/api/graphql", content=subscription_query, headers={"Content-Type": "application/json"}
    )

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]

    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 1,
        "totalItems": 2,
    }

    result_subscription_ids = {subscription["subscriptionId"] for subscription in subscriptions}
    assert result_subscription_ids == {str(subscription_1.subscription_id), str(subscription_9.subscription_id)}
    assert subscriptions[0]["status"] == "TERMINATED"
    assert subscriptions[1]["status"] == "TERMINATED"


def test_subscriptions_range_filtering_on_start_date(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)

    data = get_subscriptions_query(first=1, filter_by=[{"field": "startDate", "value": "2023-05-24"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    first_subscription = response.json()["data"]["subscriptions"]["page"][0]

    first_subscription_date = datetime.datetime.fromisoformat(first_subscription["startDate"])
    higher_than_date = first_subscription_date.isoformat()
    lower_than_date = (first_subscription_date + datetime.timedelta(days=3)).isoformat()

    data = get_subscriptions_query(
        filter_by=[
            {"field": "startDate", "value": f">={higher_than_date}"},
            {"field": "startDate", "value": f"<={lower_than_date}"},
        ]
    )
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 3,
        "totalItems": 4,
    }

    for subscription in subscriptions:
        assert higher_than_date <= subscription["startDate"] <= lower_than_date


def test_subscriptions_filtering_with_invalid_filter(
    test_client, product_type_1_subscriptions_factory, generic_product_type_1
):
    # when

    subscription_ids = product_type_1_subscriptions_factory(30)

    GenericProductOneInactive, GenericProductOne = generic_product_type_1
    subscription_1 = GenericProductOne.from_subscription(subscription_ids[0])
    subscription_1 = subscription_1.from_other_lifecycle(subscription_1, SubscriptionLifecycle.TERMINATED)
    subscription_1.save()
    subscription_9 = GenericProductOne.from_subscription(subscription_ids[8])
    subscription_9 = subscription_9.from_other_lifecycle(subscription_9, SubscriptionLifecycle.TERMINATED)
    subscription_9.save()

    data = get_subscriptions_query(
        filter_by=[
            {"field": "status", "value": SubscriptionLifecycle.TERMINATED},
            {"field": "test", "value": "invalid"},
        ]
    )
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    errors = result["errors"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert errors == [
        {
            "message": (
                "Invalid filter arguments (invalid_filters=['test'] "
                "valid_filter_keys=['customerId', 'description', 'endDate', 'insync', "
                "'note', 'product', 'productId', 'startDate', 'status', 'subscriptionId', 'tag'])"
            ),
            "path": ["subscriptions"],
            "extensions": {"error_type": "internal_error"},
        }
    ]
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 1,
        "totalItems": 2,
    }

    for subscription in subscriptions:
        assert subscription["status"] == "TERMINATED"


@pytest.mark.parametrize(
    "query_args",
    [
        lambda sid: {"filter_by": [{"field": "subscriptionId", "value": sid}]},
        lambda sid: {"query_string": f"subscription_id:{sid}"},
        lambda sid: {"query_string": f"subscriptionId:{sid}"},
        lambda sid: {"query_string": f"{sid}"},
        lambda sid: {"query_string": f"{sid.split('-')[0]}"},
        lambda sid: {"filter_by": [{"field": "customerId", "value": CUSTOMER_ID.split("-")[0]}]},
        lambda sid: {"query_string": CUSTOMER_ID.split("-")[0]},
    ],
)
def test_single_subscription(test_client, product_type_1_subscriptions_factory, generic_product_type_1, query_args):
    # given

    _, GenericProductOne = generic_product_type_1
    subscription_ids = product_type_1_subscriptions_factory(30)

    do_refresh_subscriptions_search_view()

    subscription_id = subscription_ids[10]
    subscription = db.session.execute(
        select(SubscriptionTable).filter(SubscriptionTable.subscription_id == subscription_id)
    ).scalar_one_or_none()
    subscription.customer_id = CUSTOMER_ID
    subscription.customer_descriptions = [
        SubscriptionCustomerDescriptionTable(
            subscription_id=subscription_id, customer_id=CUSTOMER_ID, description="customer alias"
        )
    ]
    db.session.add(subscription)
    db.session.commit()

    do_refresh_subscriptions_search_view()

    # when
    data = get_subscriptions_query(**query_args(subscription_id))
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    subscription = GenericProductOne.from_subscription(subscription_id)

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert subscriptions[0]["status"] == SubscriptionLifecycle.ACTIVE.name
    assert subscriptions[0]["customerDescriptions"] == [
        {
            "subscriptionId": subscription_id,
            "customerId": "a6274018-0580-41cd-9fe8-17b946706e9f",
            "description": "customer alias",
        }
    ]
    assert subscriptions[0]["productBlockInstances"] == [
        {
            "id": 0,
            "subscriptionInstanceId": str(subscription.pb_1.subscription_instance_id),
            "ownerSubscriptionId": subscription_id,
            "parent": None,
            "productBlockInstanceValues": [
                {"field": "name", "value": "PB_1"},
                {"field": "label", "value": None},
                {"field": "rt1", "value": "Value1"},
            ],
            "inUseByRelations": [],
        },
        {
            "id": 1,
            "subscriptionInstanceId": str(subscription.pb_2.subscription_instance_id),
            "ownerSubscriptionId": subscription_id,
            "parent": None,
            "productBlockInstanceValues": [
                {"field": "name", "value": "PB_2"},
                {"field": "label", "value": None},
                {"field": "rt2", "value": 42},
                {"field": "rt3", "value": "Value2"},
            ],
            "inUseByRelations": [],
        },
    ]


@pytest.mark.parametrize(
    "query_args,total_items",
    [
        (lambda ids: {}, 33),
        (lambda ids: {"filter_by": [{"field": "subscriptionId", "value": f"{ids[0]}|{ids[1]}"}]}, 2),
        (lambda ids: {"filter_by": [{"field": "subscriptionId", "value": "|".join(ids[3:8])}]}, 5),
        (lambda ids: {"query_string": f"subscriptionId:({ids[2]}|{ids[3]})"}, 2),
    ],
)
def test_multiple_subscriptions(
    test_client, product_type_1_subscriptions_factory, generic_product_type_1, query_args, total_items
):
    # given

    _, GenericProductOne = generic_product_type_1
    subscription_ids = product_type_1_subscriptions_factory(30)

    do_refresh_subscriptions_search_view()

    # when
    data = get_subscriptions_query(first=100, **query_args(subscription_ids))
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert len(subscriptions) == total_items
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": total_items - 1,
        "totalItems": total_items,
    }


@pytest.mark.parametrize(
    "query_args",
    [
        lambda sid: {"filter_by": [{"field": "subscriptionId", "value": sid}]},
        lambda sid: {"query_string": sid},
    ],
)
def test_single_subscription_with_processes(
    fastapi_app_graphql,
    test_client,
    product_type_1_subscriptions_factory,
    mocked_processes,
    mocked_processes_resumeall,  # noqa: F811
    generic_subscription_2,  # noqa: F811
    generic_subscription_1,
    query_args,
):
    # when

    product_type_1_subscriptions_factory(30)
    subscription_id = generic_subscription_1

    do_refresh_subscriptions_search_view()

    data = get_subscriptions_query_with_relations(**query_args(subscription_id))
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert subscriptions[0]["processes"]["page"][0]["processId"] == str(mocked_processes[0])


def test_single_subscription_with_depends_on_subscriptions(
    fastapi_app_graphql,
    test_client,
    product_type_1_subscriptions_factory,
    sub_one_subscription_1,
    sub_two_subscription_1,
    product_sub_list_union_subscription_1,
):
    # when

    product_type_1_subscriptions_factory(30)

    do_refresh_subscriptions_search_view()

    subscription_id = str(product_sub_list_union_subscription_1)
    data = get_subscriptions_query_with_relations(query_string=subscription_id)
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    expected_depends_on_ids = {
        str(subscription.subscription_id) for subscription in [sub_one_subscription_1, sub_two_subscription_1]
    }
    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert len(subscriptions[0]["processes"]["page"]) == 0
    depends_on_ids = subscriptions[0]["dependsOnSubscriptions"]["page"]
    result_depends_on_ids = {subscription["subscriptionId"] for subscription in depends_on_ids}
    assert result_depends_on_ids == expected_depends_on_ids
    assert len(subscriptions[0]["inUseBySubscriptions"]["page"]) == 0


def test_single_subscription_with_in_use_by_subscriptions(
    fastapi_app_graphql,
    test_client,
    product_type_1_subscriptions_factory,
    sub_one_subscription_1,
    product_sub_list_union_subscription_1,
):
    # when

    product_type_1_subscriptions_factory(30)

    subscription_id = str(sub_one_subscription_1.subscription_id)
    subscription_query = get_subscriptions_query_with_relations(
        filter_by=[{"field": "subscriptionId", "value": subscription_id}]
    )

    response = test_client.post(
        "/api/graphql", content=subscription_query, headers={"Content-Type": "application/json"}
    )

    expected_in_use_by_ids = [str(product_sub_list_union_subscription_1)]

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert len(subscriptions[0]["processes"]["page"]) == 0
    assert len(subscriptions[0]["dependsOnSubscriptions"]["page"]) == 0
    result_in_use_by_ids = [
        subscription["subscriptionId"] for subscription in subscriptions[0]["inUseBySubscriptions"]["page"]
    ]
    assert result_in_use_by_ids == expected_in_use_by_ids

    list_sub = SubscriptionModel.from_subscription(product_sub_list_union_subscription_1)
    assert subscriptions[0]["productBlockInstances"] == [
        {
            "ownerSubscriptionId": subscription_id,
            "inUseByRelations": [
                {
                    "subscription_id": str(product_sub_list_union_subscription_1),
                    "subscription_instance_id": str(list_sub.test_block.subscription_instance_id),
                }
            ],
        }
    ]


def test_single_subscription_schema(
    fastapi_app_graphql,
    test_client,
    product_type_1_subscriptions_factory,
    sub_one_subscription_1,
    sub_two_subscription_1,
    product_sub_list_union_subscription_1,
):
    # when

    product_type_1_subscriptions_factory(30)
    subscription_id = str(product_sub_list_union_subscription_1)
    data = get_subscriptions_product_block_json_schema_query(
        filter_by=[{"field": "subscriptionId", "value": subscription_id}]
    )
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]

    assert "errors" not in result
    expected = {
        "$defs": {
            "ProductBlockWithListUnionForTestInactive": {
                "description": "Valid for statuses: all others\n\nSee `active` version.",
                "properties": {
                    "name": {"anyOf": [{"type": "string"}, {"type": "null"}], "title": "Name"},
                    "subscription_instance_id": {
                        "format": "uuid",
                        "title": "Subscription Instance Id",
                        "type": "string",
                    },
                    "owner_subscription_id": {"format": "uuid", "title": "Owner Subscription Id", "type": "string"},
                    "label": {"anyOf": [{"type": "string"}, {"type": "null"}], "default": None, "title": "Label"},
                    "list_union_blocks": {
                        "items": {
                            "anyOf": [
                                {"$ref": "#/$defs/SubBlockTwoForTestInactive"},
                                {"$ref": "#/$defs/SubBlockOneForTestInactive"},
                            ]
                        },
                        "title": "List Union Blocks",
                        "type": "array",
                    },
                    "int_field": {
                        "anyOf": [{"type": "integer"}, {"type": "null"}],
                        "default": None,
                        "title": "Int Field",
                    },
                    "str_field": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "Str Field",
                    },
                    "list_field": {"items": {"type": "integer"}, "title": "List Field", "type": "array"},
                },
                "required": ["name", "subscription_instance_id", "owner_subscription_id", "list_union_blocks"],
                "title": "ProductBlockWithListUnionForTestInactive",
                "type": "object",
            },
            "ProductLifecycle": {
                "enum": ["active", "pre production", "phase out", "end of life"],
                "title": "ProductLifecycle",
                "type": "string",
            },
            "ProductModel": {
                "description": "Represent the product as defined in the database as a dataclass.",
                "properties": {
                    "product_id": {"format": "uuid", "title": "Product Id", "type": "string"},
                    "name": {"title": "Name", "type": "string"},
                    "description": {"title": "Description", "type": "string"},
                    "product_type": {"title": "Product Type", "type": "string"},
                    "tag": {"title": "Tag", "type": "string"},
                    "status": {"$ref": "#/$defs/ProductLifecycle"},
                    "created_at": {
                        "anyOf": [{"format": "date-time", "type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "Created At",
                    },
                    "end_date": {
                        "anyOf": [{"format": "date-time", "type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "End Date",
                    },
                },
                "required": ["product_id", "name", "description", "product_type", "tag", "status"],
                "title": "ProductModel",
                "type": "object",
            },
            "SubBlockOneForTestInactive": {
                "description": "Valid for statuses: all others\n\nSee `active` version.",
                "properties": {
                    "name": {"anyOf": [{"type": "string"}, {"type": "null"}], "title": "Name"},
                    "subscription_instance_id": {
                        "format": "uuid",
                        "title": "Subscription Instance Id",
                        "type": "string",
                    },
                    "owner_subscription_id": {"format": "uuid", "title": "Owner Subscription Id", "type": "string"},
                    "label": {"anyOf": [{"type": "string"}, {"type": "null"}], "default": None, "title": "Label"},
                    "int_field": {
                        "anyOf": [{"type": "integer"}, {"type": "null"}],
                        "default": None,
                        "title": "Int Field",
                    },
                    "str_field": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "Str Field",
                    },
                },
                "required": ["name", "subscription_instance_id", "owner_subscription_id"],
                "title": "SubBlockOneForTestInactive",
                "type": "object",
            },
            "SubBlockTwoForTestInactive": {
                "description": "Valid for statuses: all others\n\nSee `active` version.",
                "properties": {
                    "name": {"anyOf": [{"type": "string"}, {"type": "null"}], "title": "Name"},
                    "subscription_instance_id": {
                        "format": "uuid",
                        "title": "Subscription Instance Id",
                        "type": "string",
                    },
                    "owner_subscription_id": {"format": "uuid", "title": "Owner Subscription Id", "type": "string"},
                    "label": {"anyOf": [{"type": "string"}, {"type": "null"}], "default": None, "title": "Label"},
                    "int_field_2": {"title": "Int Field 2", "type": "integer"},
                },
                "required": ["name", "subscription_instance_id", "owner_subscription_id", "int_field_2"],
                "title": "SubBlockTwoForTestInactive",
                "type": "object",
            },
            "SubscriptionLifecycle": {
                "enum": ["initial", "active", "migrating", "disabled", "terminated", "provisioning"],
                "title": "SubscriptionLifecycle",
                "type": "string",
            },
        },
        "description": "Valid for statuses: all others\n\nSee `active` version.",
        "properties": {
            "product": {"$ref": "#/$defs/ProductModel"},
            "customer_id": {"title": "Customer Id", "type": "string"},
            "subscription_id": {"format": "uuid", "title": "Subscription Id", "type": "string"},
            "description": {"default": "Initial subscription", "title": "Description", "type": "string"},
            "status": {"allOf": [{"$ref": "#/$defs/SubscriptionLifecycle"}], "default": "initial"},
            "insync": {"default": False, "title": "Insync", "type": "boolean"},
            "start_date": {
                "anyOf": [{"format": "date-time", "type": "string"}, {"type": "null"}],
                "default": None,
                "title": "Start Date",
            },
            "end_date": {
                "anyOf": [{"format": "date-time", "type": "string"}, {"type": "null"}],
                "default": None,
                "title": "End Date",
            },
            "note": {"anyOf": [{"type": "string"}, {"type": "null"}], "default": None, "title": "Note"},
            "test_block": {"anyOf": [{"$ref": "#/$defs/ProductBlockWithListUnionForTestInactive"}, {"type": "null"}]},
        },
        "required": ["product", "customer_id", "test_block"],
        "title": "ProductSubListUnionInactive",
        "type": "object",
    }
    actual = subscriptions[0]["_schema"]
    assert actual == expected


def test_single_subscription_metadata_and_schema(
    fastapi_app_graphql,
    test_client,
    sub_one_subscription_1,
):
    # when
    expected_metadata = {"some_metadata_prop": ["test value 1", "test 2"]}
    subscription_id = str(sub_one_subscription_1.subscription_id)
    subscription_metadata = SubscriptionMetadataTable(subscription_id=subscription_id, metadata_=expected_metadata)
    db.session.add(subscription_metadata)
    db.session.commit()

    data = get_subscriptions_with_metadata_and_schema_query(
        filter_by=[{"field": "subscriptionId", "value": subscription_id}]
    )
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]

    assert "errors" not in result
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert subscriptions[0]["metadata"] == expected_metadata
    assert subscriptions[0]["_metadataSchema"] == {
        "title": "Metadata",
        "type": "object",
        "properties": {
            "some_metadata_prop": {"title": "Some Metadata Prop", "type": "array", "items": {"type": "string"}}
        },
        "required": ["some_metadata_prop"],
    }


def test_subscriptions_product_generic_one(
    fastapi_app_graphql,
    test_client,
    product_type_1_subscriptions_factory,
):
    # when

    subscriptions = product_type_1_subscriptions_factory(30)
    subscription_id = str(subscriptions[0])
    subscription = db.session.execute(
        select(SubscriptionTable).filter(SubscriptionTable.subscription_id == subscription_id)
    ).scalar_one_or_none()
    subscription.customer_descriptions = [
        SubscriptionCustomerDescriptionTable(customer_id=CUSTOMER_ID, description="customer alias")
    ]
    db.session.add(subscription)
    db.session.commit()

    data = get_subscriptions_product_generic_one(filter_by=[{"field": "subscriptionId", "value": subscription_id}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert subscriptions[0]["status"] == SubscriptionLifecycle.ACTIVE.name
    assert subscriptions[0]["pb1"] == {"rt1": "Value1"}
    assert subscriptions[0]["pb2"] == {"rt2": 42, "rt3": "Value2"}
    assert subscriptions[0]["customerDescriptions"] == [
        {
            "subscriptionId": subscription_id,
            "customerId": "a6274018-0580-41cd-9fe8-17b946706e9f",
            "description": "customer alias",
        }
    ]


def test_single_subscription_product_list_union_type(
    fastapi_app_graphql,
    test_client,
    product_sub_list_union_subscription_1,
):
    # when

    subscription_id = str(product_sub_list_union_subscription_1)
    data = get_subscriptions_product_sub_list_union(filter_by=[{"field": "subscriptionId", "value": subscription_id}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert subscriptions[0]["status"] == SubscriptionLifecycle.ACTIVE.name
    assert subscriptions[0]["testBlock"] == {
        "intField": 1,
        "strField": "blah",
        "listField": [2],
        "listUnionBlocks": [{"intField2": 3}, {"intField": 1, "strField": "blah"}],
    }


def test_single_subscription_product_list_union_type_provisioning_subscription(
    fastapi_app_graphql,
    test_client,
    product_sub_list_union_subscription_1,
):
    # when

    subscription = SubscriptionModel.from_subscription(product_sub_list_union_subscription_1)
    subscription = subscription.from_other_lifecycle(subscription, SubscriptionLifecycle.PROVISIONING)
    subscription.save()

    subscription_id = str(product_sub_list_union_subscription_1)
    data = get_subscriptions_product_sub_list_union(filter_by=[{"field": "subscriptionId", "value": subscription_id}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert subscriptions[0]["testBlock"] == {
        "intField": 1,
        "strField": "blah",
        "listField": [2],
        "listUnionBlocks": [{"intField2": 3}, {"intField": 1, "strField": "blah"}],
    }


def test_single_subscription_product_list_union_type_terminated_subscription(
    fastapi_app_graphql,
    test_client,
    product_sub_list_union_subscription_1,
):
    # when

    subscription = SubscriptionModel.from_subscription(product_sub_list_union_subscription_1)
    subscription = subscription.from_other_lifecycle(subscription, SubscriptionLifecycle.TERMINATED)
    subscription.save()

    subscription_id = str(product_sub_list_union_subscription_1)
    data = get_subscriptions_product_sub_list_union(filter_by=[{"field": "subscriptionId", "value": subscription_id}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert subscriptions[0]["testBlock"] == {
        "intField": 1,
        "strField": "blah",
        "listField": [2],
        "listUnionBlocks": [{"intField2": 3}, {"intField": 1, "strField": "blah"}],
    }



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ipaddress import IPv4Address, IPv6Address
from uuid import UUID

from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle


class TestProductBlockInactive(ProductBlockModel, product_block_name="Test Product Block"):
    an_int: int | None = None
    a_str: str | None = None
    a_bool: bool | None = None
    an_uuid: UUID | None = None
    an_ipv4: IPv4Address | None = None
    an_ipv6: IPv6Address | None = None


class TestProductBlockProvisioning(TestProductBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    an_int: int
    a_str: str
    a_bool: bool | None = None
    an_uuid: UUID
    an_ipv4: IPv4Address | None = None
    an_ipv6: IPv6Address | None = None


class TestProductBlock(TestProductBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    an_int: int
    a_str: str
    a_bool: bool
    an_uuid: UUID
    an_ipv4: IPv4Address
    an_ipv6: IPv6Address



from collections.abc import Iterable

from sqlalchemy import Select, select

from orchestrator.db import WorkflowTable, db
from orchestrator.schemas import StepSchema, WorkflowSchema
from orchestrator.workflows import get_workflow


def _get_steps(workflow: WorkflowTable) -> list[StepSchema]:
    if registered_workflow := get_workflow(workflow.name):
        return [StepSchema(name=step.name) for step in registered_workflow.steps]
    raise AssertionError(f"Workflow {workflow.name} should be registered")


def _to_workflow_schema(workflow: WorkflowTable, include_steps: bool = False) -> WorkflowSchema:
    extra_kwargs = {"steps": _get_steps(workflow)} if include_steps else {}

    return WorkflowSchema(
        workflow_id=workflow.workflow_id,
        name=workflow.name,
        target=workflow.target,
        description=workflow.description,
        created_at=workflow.created_at,
        **extra_kwargs,
    )


def get_workflows(
    filters: dict | None = None, include_steps: bool = False, include_deleted: bool = False
) -> Iterable[WorkflowSchema]:
    def _add_filter(stmt: Select) -> Select:
        for k, v in (filters or {}).items():
            stmt = stmt.where(WorkflowTable.__dict__[k] == v)
        return stmt

    stmt = select(WorkflowTable) if include_deleted else WorkflowTable.select()
    workflows = db.session.scalars(_add_filter(stmt)).all()

    return [_to_workflow_schema(wf, include_steps=include_steps) for wf in workflows]


def get_workflow_by_name(workflow_name: str) -> WorkflowTable | None:
    return db.session.scalar(select(WorkflowTable).where(WorkflowTable.name == workflow_name))



from typing import Union

from orchestrator.types import is_of_type


def test_is_of_type():
    """Some tests to see type checks are valid."""
    assert is_of_type(int, Union[int, str])
    assert is_of_type(int, Union[str, int])
    assert is_of_type(str, Union[int, str])
    assert is_of_type(str, Union[str, int])
    assert is_of_type(list[str], Union[str, int]) is False



from collections import defaultdict
from uuid import UUID

import structlog
from more_itertools import flatten
from sqlalchemy import func, select
from sqlalchemy.sql.expression import Delete, Insert, Update
from sqlalchemy.sql.selectable import ScalarSelect

from orchestrator.cli.domain_gen_helpers.helpers import sql_compile
from orchestrator.cli.domain_gen_helpers.product_block_helpers import get_product_block_id, get_product_block_ids
from orchestrator.cli.domain_gen_helpers.types import DomainModelChanges
from orchestrator.cli.helpers.input_helpers import _enumerate_menu_keys, _prompt_user_menu, get_user_input
from orchestrator.cli.helpers.print_helpers import COLOR, noqa_print, print_fmt, str_fmt
from orchestrator.db import db
from orchestrator.db.models import (
    ProductBlockTable,
    ProductTable,
    ResourceTypeTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    SubscriptionTable,
    product_block_resource_type_association,
)
from orchestrator.domain.base import ProductBlockModel

logger = structlog.get_logger(__name__)


def get_resource_type(resource_type: str) -> ScalarSelect:
    return get_resource_types([resource_type])


def get_resource_types(resource_types: list[str] | set[str]) -> ScalarSelect:
    return (
        select(ResourceTypeTable.resource_type_id)
        .where(ResourceTypeTable.resource_type.in_(resource_types))
        .scalar_subquery()
    )


def map_create_resource_types(resource_type_names: set[str], updated_resource_types: set[str]) -> set[str]:
    """Map resource types to create.

    Args:
        resource_type_names: List of resource type names.
        updated_resource_types: List of resource types that get renamed and don't need to be created.

    Returns: List of resource type names that can be created.
    """
    existing_resource_types = set(
        db.session.scalars(
            select(ResourceTypeTable.resource_type).where(ResourceTypeTable.resource_type.in_(resource_type_names))
        )
    )
    existing_resource_types.update(updated_resource_types)
    return {rt for rt in resource_type_names if rt not in existing_resource_types}


def find_resource_within_blocks(
    resource_type_names: list[str], product_blocks: dict[str, type[ProductBlockModel]]
) -> set[str]:
    """Find resource types within product blocks.

    Args:
        resource_type_names: List of resource type names.
        product_blocks: Dict of product blocks mapped by product block name, used to check if the resource type still exists in a product block that hasn't changed.

    Returns: List of the resource_type_names that are found within product blocks.
    """
    keys = flatten([product_block._non_product_block_fields_.keys() for product_block in product_blocks.values()])
    return {field_name for field_name in keys if field_name in resource_type_names}


def rename_resource_type_inputs(
    resource_type_choices: dict[str, set[str]],
    renamed_resource_types: dict[str, str],
) -> tuple[str | None, str | None]:
    old_resource_types = set(resource_type_choices.keys())
    noqa_print("Which resource type would you want to rename?")
    old_rt = _prompt_user_menu(
        [*[(p, p) for p in old_resource_types], ("continue", None)],
        keys=[*_enumerate_menu_keys(old_resource_types), "q"],
    )
    if not old_rt:
        return None, None

    choices = resource_type_choices[old_rt] - set(renamed_resource_types.values())
    noqa_print(f"\nWhich resource type should rename {old_rt}?")
    new_rt = _prompt_user_menu(
        [*[(p, p) for p in choices], ("cancel", None)],
        keys=[*_enumerate_menu_keys(choices), "q"],
    )
    return old_rt, new_rt if new_rt else None


def map_rename_resource_types(
    block_diffs: dict[str, dict[str, set[str]]],
    product_blocks: dict[str, type[ProductBlockModel]],
) -> dict[str, str]:
    """Map resource types to rename.

    Args:
        block_diffs: Dict with product block differences.
            - key: product block name
            - value: Dict with differences between model and database.
                - key: difference name, 'missing_resource_types_in_model' and 'missing_resource_types_in_db' are used to check if a resource type can be renamed.
                - value: Set of resource type names.
        product_blocks: Dict of product blocks mapped by product block name, used to check if the resource type still exists in a product block.

    Returns: Dict with resource types that can be updated.
        - key: old resource type name.
        - value: new resource type name.
    """

    renamed_resource_types: dict[str, str] = {}
    new_resource_types: set[str] = set()
    existing_resource_types = list(db.session.scalars(select(ResourceTypeTable.resource_type)))
    possible_rt_choices: dict[str, set[str]] = defaultdict(set)

    for diff in block_diffs.values():
        db_resource_types = list(diff.get("missing_resource_types_in_model", set()))
        model_resource_types = list(diff.get("missing_resource_types_in_db", set()))
        existing_in_blocks = find_resource_within_blocks(db_resource_types, product_blocks)

        old_block_resource_types = {
            resource_types for resource_types in db_resource_types if resource_types not in existing_in_blocks
        }
        new_block_resource_types = {
            resource_type for resource_type in model_resource_types if resource_type not in existing_resource_types
        }

        new_resource_types.update(new_block_resource_types)
        for rt in old_block_resource_types:
            possible_rt_choices[rt].update(new_block_resource_types)

    print_fmt("\nRename resource types", flags=[COLOR.BOLD, COLOR.UNDERLINE])
    while len(possible_rt_choices.keys()) > 0 and len(new_resource_types) > 0:
        old_rt, new_rt = rename_resource_type_inputs(possible_rt_choices, renamed_resource_types)
        if not old_rt:
            break
        if not new_rt:
            continue

        possible_rt_choices = {
            key: {value for value in values if value != new_rt}
            for key, values in possible_rt_choices.items()
            if key != old_rt
        }
        new_resource_types.remove(new_rt)
        renamed_resource_types[old_rt] = new_rt
    return renamed_resource_types


def update_block_resource_type_input(
    old_props: list[str], new_props: list[str], block_name: str
) -> tuple[str | None, str | None]:
    noqa_print(f"Which resource type would you want to update in {block_name} Block?")
    old_rt = _prompt_user_menu(
        [*[(p, p) for p in old_props], ("continue", None)],
        keys=[*_enumerate_menu_keys(old_props), "q"],
    )
    if not old_rt:
        return None, None

    noqa_print(f"\nWhich resource type should update {old_rt}?")
    new_rt = _prompt_user_menu(
        [*[(p, p) for p in new_props], ("cancel", None)],
        keys=[*_enumerate_menu_keys(new_props), "q"],
    )
    return old_rt, new_rt if new_rt else None


def map_update_product_block_resource_types(
    block_diffs: dict[str, dict[str, set[str]]],
    renamed_resource_types: dict[str, str],
) -> dict[str, dict[str, str]]:
    """Map resource types to update per product block.

    Args:
        block_diffs: Dict with product block differences.
            - key: product block name.
            - value: Dict with differences between model and database.
                - key: difference name, 'missing_resource_types_in_model' and 'missing_resource_types_in_db' are used to check if a resource type can be renamed.
                - value: Set of resource type names.
        renamed_resource_types: Dict of renamed resource types, old name as key and new name as value.

    Returns: Dict with resource types per product block that can be updated.
        - key: product block name.
        - value: Dict with resource types to update.
            - key: old resource type name.
            - value: new resource type name.
    """

    updates = {}
    updated_new_rts = renamed_resource_types.values()
    print_fmt("\nUpdate block resource types", flags=[COLOR.BOLD, COLOR.UNDERLINE])
    for block_name, diff in block_diffs.items():
        db_props = list(diff.get("missing_resource_types_in_model", set()))
        model_props = list(diff.get("missing_resource_types_in_db", set()))

        old_rts = [prop for prop in db_props if prop not in renamed_resource_types]
        new_rts = [prop for prop in model_props if prop not in updated_new_rts]
        while len(old_rts) > 0 and len(new_rts) > 0:
            old_rt, new_rt = update_block_resource_type_input(old_rts, new_rts, block_name)
            if not old_rt:
                break
            if not new_rt:
                continue
            old_rts = [rt for rt in old_rts if rt != old_rt]
            new_rts = [rt for rt in new_rts if rt != new_rt]
            updates[block_name] = {old_rt: new_rt}
    return updates


def map_delete_resource_types(
    resource_types: dict[str, set[str]],
    updated_resource_types: list[str],
    product_blocks: dict[str, type[ProductBlockModel]],
) -> set[str]:
    """Map resource types to delete.

    Args:
        resource_types: List of resource type names.
        updated_resource_types: List of resource types that get renamed and shouldn't be deleted.
        product_blocks: Dict of product blocks mapped by product block name, used to check if the resource type still exists in a product block.

    Returns: List of resource type names that can be deleted.
    """
    resource_type_names = resource_types.keys()
    stmt = select(ResourceTypeTable.resource_type).where(ResourceTypeTable.resource_type.in_(resource_type_names))
    existing_resource_types = db.session.scalars(stmt).all()
    existing_names = [r_type for r_type in existing_resource_types if r_type in resource_type_names]
    rt_with_existing_instances = {
        *find_resource_within_blocks(existing_names, product_blocks),
        *updated_resource_types,
    }
    return {name for name in existing_names if name not in rt_with_existing_instances}


def get_product_instance_count(product_id: UUID) -> int:
    return (
        db.session.scalar(
            select(func.count()).select_from(SubscriptionTable).where(SubscriptionTable.product_id == product_id)
        )
        or 0
    )


def get_block_instance_count(product_block_id: UUID) -> int:
    stmt = (
        select(func.count())
        .select_from(SubscriptionInstanceTable)
        .where(SubscriptionInstanceTable.product_block_id == product_block_id)
    )
    return db.session.scalar(stmt) or 0


def _has_product_existing_instances(product_name: str) -> bool:
    stmt = select(ProductTable).where(ProductTable.name == product_name)
    product: ProductTable | None = db.session.scalars(stmt).first()
    return bool(product and get_product_instance_count(product.product_id))


def _find_new_relations(block_name: str, relations: dict[str, set[str]]) -> set[str]:
    return set(flatten((list(v) for k, v in relations.items() if block_name in k)))


def map_create_resource_type_instances(changes: DomainModelChanges) -> dict[str, set[str]]:
    """Map resource types that need a default value.

    Resource types need a default value when the related product block is used in an existing instance or will be used in an existing instance.

    Args:
        changes: DomainModelChanges class with all changes.

    Returns: Dict with resource types that need a default value.
        - key: resource type name.
        - value: set of product block names.
    """

    def _has_existing_instances(block_name: str) -> bool:
        stmt = select(ProductBlockTable).where(ProductBlockTable.name == block_name)
        block: ProductBlockTable | None = db.session.scalars(stmt).first()
        if block and get_block_instance_count(block.product_block_id):
            return True

        related_block_names = _find_new_relations(block_name, changes.create_product_block_relations)
        if related_block_names:
            return any(_has_existing_instances(name) for name in related_block_names)

        related_product_names = _find_new_relations(block_name, changes.create_product_to_block_relations)
        return any(_has_product_existing_instances(name) for name in related_product_names)

    return {
        resource_name: {name for name in block_names if _has_existing_instances(name)}
        for resource_name, block_names in changes.create_resource_type_relations.items()
    }


def generate_create_resource_types_sql(
    resource_types: set[str], inputs: dict[str, dict[str, str]], revert: bool = False
) -> list[str]:
    """Generate SQL to create resource types.

    Args:
        resource_types: List of resource type names.
        inputs: Optional Dict to add default value to the resource type for existing product block instances.
            - key: Resource type name.
            - value: Dict with product block by default value.
                - key: Product block name.
                - value: Default value for the resource type.
        revert: Revert bool

    Returns: List of SQL strings to create resource type.
    """
    print_fmt("\nCreate resource types", flags=[COLOR.BOLD, COLOR.UNDERLINE])

    def create_resource_type(resource_type: str) -> str:
        if revert:
            stmt = select(ResourceTypeTable.description).where(ResourceTypeTable.resource_type == resource_type)
            description = db.session.scalars(stmt).one()
        else:
            description = inputs.get(resource_type, {}).get("description") or get_user_input(
                f"Supply description for new resource type {str_fmt(resource_type, flags=[COLOR.MAGENTA])}: "
            )

        return sql_compile(
            Insert(ResourceTypeTable).values({"resource_type": resource_type, "description": description})
        )

    return [create_resource_type(resource_type) for resource_type in resource_types]


def generate_rename_resource_types_sql(resource_types: dict[str, str]) -> list[str]:
    """Generate SQL to update resource types.

    Args:
        resource_types: Dict with new resource type name by old resource type name.
            - key: old resource type name.
            - value: new resource type name.

    Returns: List of SQL strings to update resource types.
    """

    def update_resource_type(old_rt_name: str, new_rt_name: str) -> str:
        return sql_compile(
            Update(ResourceTypeTable)
            .where(ResourceTypeTable.resource_type == old_rt_name)
            .values(resource_type=new_rt_name)
        )

    return [update_resource_type(*item) for item in resource_types.items()]


def generate_delete_resource_types_sql(resource_types: set[str]) -> list[str]:
    """Generate SQL to delete resource types.

    Args:
        resource_types: List of resource type names.

    Returns: List of SQL strings to delete resource types.
    """
    if not resource_types:
        return []
    return [
        sql_compile(
            Delete(SubscriptionInstanceValueTable).where(
                SubscriptionInstanceValueTable.resource_type_id.in_(get_resource_types(resource_types))
            )
        ),
        sql_compile(Delete(ResourceTypeTable).where(ResourceTypeTable.resource_type.in_(resource_types))),
    ]


def generate_create_resource_type_relations_sql(resource_types: dict[str, set[str]]) -> list[str]:
    """Generate SQL to create resource type relations.

    Args:
        resource_types: Dict with product blocks by resource type
            - key: Resource type name.
            - value: Set of product block names to relate with.

    Returns: List of SQL strings to create relation between product blocks and resource type.
    """

    def create_resource_type_relation(resource_type: str, block_names: set[str]) -> str:
        resource_type_id_sql = get_resource_type(resource_type)

        def create_block_relation_dict(block_name: str) -> dict[str, str | ScalarSelect]:
            block_id_sql = get_product_block_id(block_name)
            return {"resource_type_id": resource_type_id_sql, "product_block_id": block_id_sql}

        block_relation_dicts = [create_block_relation_dict(block_name) for block_name in block_names]
        return sql_compile(Insert(product_block_resource_type_association).values(block_relation_dicts))

    return [create_resource_type_relation(*item) for item in resource_types.items()]


def generate_create_resource_type_instance_values_sql(
    resource_types: dict[str, set[str]], inputs: dict[str, dict[str, str]]
) -> list[str]:
    """Generate SQL to create resource type instance values for existing instances.

    Args:
        resource_types: Dict with product blocks by resource type
            - key: Resource type name.
            - value: Set of product block names to relate with.
        inputs: Optional Dict to add default value to the resource type for existing product block instances.
            - key: Resource type name.
            - value: Dict with product block by default value.
                - key: Product block name.
                - value: Default value for the resource type.

    Returns: List of SQL strings to create subscription instance values for existing product block instances.
    """
    print_fmt("\nCreate subscription instance values", flags=[COLOR.BOLD, COLOR.UNDERLINE])

    def map_subscription_instance_relations(resource_type: str, block_name: str) -> str:
        input_value = inputs.get(resource_type, {}).get(block_name) or inputs.get(resource_type, {}).get("value")
        value = input_value or get_user_input(
            f"Supply default subscription instance value for resource type {str_fmt(resource_type, flags=[COLOR.MAGENTA])} and product block {str_fmt(block_name, flags=[COLOR.BOLD])}: "
        )

        query = """
                WITH subscription_instance_ids AS (
                    SELECT subscription_instances.subscription_instance_id
                    FROM   subscription_instances
                    WHERE  subscription_instances.product_block_id IN (
                        SELECT product_blocks.product_block_id
                        FROM   product_blocks
                        WHERE  product_blocks.name = '{0}'
                    )
                )

                INSERT INTO
                    subscription_instance_values (subscription_instance_id, resource_type_id, value)
                SELECT
                    subscription_instance_ids.subscription_instance_id,
                    resource_types.resource_type_id,
                    '{2}'
                FROM resource_types
                CROSS JOIN subscription_instance_ids
                WHERE resource_types.resource_type = '{1}'
        """
        sql_string = query.format(block_name, resource_type, value)
        logger.debug("Generated SQL", sql_string=sql_string)
        return sql_string

    return [
        sql
        for resource_type, block_names in resource_types.items()
        for block_name in block_names
        if (sql := map_subscription_instance_relations(resource_type, block_name))
    ]


def generate_delete_resource_type_relations_sql(delete_resource_types: dict[str, set[str]]) -> list[str]:
    """Generate SQL to delete resource type relations and its instance values.

    Args:
        delete_resource_types: Dict with product blocks by resource type
            - key: Resource type name.
            - value: Set of product block names to relate with.

    Returns: List of SQL strings to delete relations between product blocks and resource type.
    """

    def delete_resource_type_relation(resource_type: str, block_names: set[str]) -> list[str]:
        block_ids_sql = get_product_block_ids(block_names)
        resource_type_id_sql = get_resource_type(resource_type)
        subscription_instance_id_sql = (
            select(SubscriptionInstanceTable.subscription_instance_id)
            .where(SubscriptionInstanceTable.subscription_instance_id.in_(block_ids_sql))
            .scalar_subquery()
        )
        return [
            sql_compile(
                Delete(product_block_resource_type_association).where(
                    product_block_resource_type_association.c.product_block_id.in_(block_ids_sql),
                    product_block_resource_type_association.c.resource_type_id == resource_type_id_sql,
                )
            ),
            sql_compile(
                Delete(SubscriptionInstanceValueTable).where(
                    SubscriptionInstanceValueTable.subscription_instance_id.in_(subscription_instance_id_sql),
                    product_block_resource_type_association.c.resource_type_id == resource_type_id_sql,
                )
            ),
        ]

    return list(flatten([delete_resource_type_relation(*item) for item in delete_resource_types.items()]))


def generate_update_resource_type_block_relations_sql(block_rt_updates: dict[str, dict[str, str]]) -> list[str]:
    """Generate SQL to update resource type block relations.

    Args:
        block_rt_updates: Dict with rt updates per product block.
            - key: product block name.
            - value: Dict with new resource type name by old resource type name:
                - key: old resource type name.
                - value: new resource type name.

    Returns: List of SQL strings to update resource types.
    """

    def update_block_resource_types(block_name: str, rt_updates: dict[str, str]) -> list[str]:
        block_id_sql = get_product_block_id(block_name)

        def update_block_relation(old_rt_name: str, new_rt_name: str) -> str:
            old_rt_id = get_resource_type(old_rt_name)
            new_rt_id = get_resource_type(new_rt_name)
            return sql_compile(
                Update(product_block_resource_type_association)
                .where(
                    product_block_resource_type_association.c.product_block_id.in_(block_id_sql),
                    product_block_resource_type_association.c.resource_type_id.in_(old_rt_id),
                )
                .values(resource_type_id=new_rt_id)
            )

        return [update_block_relation(*item) for item in rt_updates.items()]

    return list(flatten([update_block_resource_types(*item) for item in block_rt_updates.items()]))


def generate_update_resource_type_instance_values_sql(block_rt_updates: dict[str, dict[str, str]]) -> list[str]:
    """Generate SQL to update resource type instance values.

    Args:
        block_rt_updates: Dict with rt updates per product block.
            - key: product block name.
            - value: Dict with new resource type name by old resource type name:
                - key: old resource type name.
                - value: new resource type name.

    Returns: List of SQL strings to update resource types.
    """

    def update_block_resource_types(block_name: str, rt_updates: dict[str, str]) -> list[str]:
        instance_ids_sql = (
            select(SubscriptionInstanceTable.subscription_instance_id)
            .where(SubscriptionInstanceTable.product_block_id.in_(get_product_block_id(block_name)))
            .scalar_subquery()
        )

        def update_instance_values(old_rt_name: str, new_rt_name: str) -> str:
            old_rt_id = get_resource_type(old_rt_name)
            new_rt_id = get_resource_type(new_rt_name)
            return sql_compile(
                Update(SubscriptionInstanceValueTable)
                .where(
                    SubscriptionInstanceValueTable.subscription_instance_id.in_(instance_ids_sql),
                    SubscriptionInstanceValueTable.resource_type_id.in_(old_rt_id),
                )
                .values(resource_type_id=new_rt_id)
            )

        return [update_instance_values(*item) for item in rt_updates.items()]

    return list(flatten([update_block_resource_types(*item) for item in block_rt_updates.items()]))



from typing import TYPE_CHECKING, Annotated

import strawberry

from orchestrator.config.assignee import Assignee
from orchestrator.db import WorkflowTable
from orchestrator.graphql.schemas.helpers import get_original_model
from orchestrator.schemas import StepSchema, WorkflowSchema
from orchestrator.workflows import get_workflow

if TYPE_CHECKING:
    from orchestrator.graphql.schemas.product import ProductType


@strawberry.experimental.pydantic.type(model=StepSchema, all_fields=True)
class Step:
    assignee: Assignee | None


@strawberry.experimental.pydantic.type(model=WorkflowSchema, all_fields=True)
class Workflow:
    @strawberry.field(description="Return all products that can use this workflow")  # type: ignore
    async def products(self) -> list[Annotated["ProductType", strawberry.lazy(".product")]]:
        from orchestrator.graphql.schemas.product import ProductType

        model = get_original_model(self, WorkflowTable)

        return [ProductType.from_pydantic(product) for product in model.products]

    @strawberry.field(description="Return all steps for this workflow")  # type: ignore
    def steps(self) -> list[Step]:
        return [Step(name=step.name, assignee=step.assignee) for step in get_workflow(self.name).steps]  # type: ignore



import pytest
from orchestrator.forms import FormValidationError
from orchestrator.types import SubscriptionLifecycle

from products.product_types.example1 import Example1
from test.unit_tests.workflows import assert_complete, extract_state, run_workflow


@pytest.mark.workflow
def test_happy_flow(responses, example1_subscription):
    # when

    # TODO: insert mocks here if needed

    result, _, _ = run_workflow("terminate_example1", [{"subscription_id": example1_subscription}, {}])

    # then

    assert_complete(result)
    state = extract_state(result)
    assert "subscription" in state

    # Check subscription in DB

    example1 = Example1.from_subscription(example1_subscription)
    assert example1.end_date is not None
    assert example1.status == SubscriptionLifecycle.TERMINATED


@pytest.mark.workflow
def test_can_only_terminate_when_modifiable_boolean_is_true(responses, example1_subscription):
    # given

    # TODO: set test conditions or fixture so that "Add an model_validator that requires some condition(s)" triggers

    # when

    with pytest.raises(FormValidationError) as error:
        result, _, _ = run_workflow("terminate_example1", [{"subscription_id": example1_subscription}, {}])

    # then

    assert error.value.errors[0]["msg"] == "Add an model_validator that requires some condition(s)"



"""rename_parent_and_child_block_relations.

Revision ID: bed6bc0b197a
Revises: 19cdd3ab86f6
Create Date: 2022-02-16 14:20:19.813435

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "bed6bc0b197a"
down_revision = "19cdd3ab86f6"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.drop_index("product_block_relation_p_c_ix", table_name="product_block_relations")
    op.drop_constraint("product_block_relations_parent_id_fkey", "product_block_relations", type_="foreignkey")
    op.drop_constraint("product_block_relations_child_id_fkey", "product_block_relations", type_="foreignkey")
    op.alter_column("product_block_relations", "parent_id", new_column_name="in_use_by_id")
    op.alter_column("product_block_relations", "child_id", new_column_name="depends_on_id")
    op.create_index(
        "product_block_relation_i_d_ix", "product_block_relations", ["in_use_by_id", "depends_on_id"], unique=True
    )
    op.create_foreign_key(
        "product_block_relations_in_use_by_id_fkey",
        "product_block_relations",
        "product_blocks",
        ["in_use_by_id"],
        ["product_block_id"],
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        "product_block_relations_depends_on_id_fkey",
        "product_block_relations",
        "product_blocks",
        ["depends_on_id"],
        ["product_block_id"],
        ondelete="CASCADE",
    )

    op.drop_index("subscription_relation_p_c_o_ix", table_name="subscription_instance_relations")
    op.drop_constraint(
        "subscription_instance_relations_parent_id_fkey", "subscription_instance_relations", type_="foreignkey"
    )
    op.drop_constraint(
        "subscription_instance_relations_child_id_fkey", "subscription_instance_relations", type_="foreignkey"
    )
    op.alter_column("subscription_instance_relations", "parent_id", new_column_name="in_use_by_id")
    op.alter_column("subscription_instance_relations", "child_id", new_column_name="depends_on_id")
    op.create_index(
        "subscription_relation_i_d_o_ix",
        "subscription_instance_relations",
        ["in_use_by_id", "depends_on_id", "order_id"],
        unique=True,
    )
    op.create_foreign_key(
        "subscription_instance_relations_in_use_by_id_fkey",
        "subscription_instance_relations",
        "subscription_instances",
        ["in_use_by_id"],
        ["subscription_instance_id"],
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        "subscription_instance_relations_depends_on_id_fkey",
        "subscription_instance_relations",
        "subscription_instances",
        ["depends_on_id"],
        ["subscription_instance_id"],
        ondelete="CASCADE",
    )


def downgrade() -> None:
    op.drop_index("product_block_relation_i_d_ix", table_name="product_block_relations")
    op.drop_constraint("product_block_relations_in_use_by_id_fkey", "product_block_relations", type_="foreignkey")
    op.drop_constraint("product_block_relations_depends_on_id_fkey", "product_block_relations", type_="foreignkey")
    op.alter_column("product_block_relations", "in_use_by_id", new_column_name="parent_id")
    op.alter_column("product_block_relations", "depends_on_id", new_column_name="child_id")
    op.create_index("product_block_relation_p_c_ix", "product_block_relations", ["parent_id", "child_id"], unique=True)
    op.create_foreign_key(
        "product_block_relations_parent_id_fkey",
        "product_block_relations",
        "product_blocks",
        ["parent_id"],
        ["product_block_id"],
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        "product_block_relations_child_id_fkey",
        "product_block_relations",
        "product_blocks",
        ["child_id"],
        ["product_block_id"],
        ondelete="CASCADE",
    )

    op.drop_index("subscription_relation_i_d_o_ix", table_name="subscription_instance_relations")
    op.drop_constraint(
        "subscription_instance_relations_in_use_by_id_fkey", "subscription_instance_relations", type_="foreignkey"
    )
    op.drop_constraint(
        "subscription_instance_relations_depends_on_id_fkey", "subscription_instance_relations", type_="foreignkey"
    )
    op.alter_column("subscription_instance_relations", "in_use_by_id", new_column_name="parent_id")
    op.alter_column("subscription_instance_relations", "depends_on_id", new_column_name="child_id")
    op.create_index(
        "subscription_relation_p_c_o_ix",
        "subscription_instance_relations",
        ["parent_id", "child_id", "order_id"],
        unique=True,
    )
    op.create_foreign_key(
        "subscription_instance_relations_parent_id_fkey",
        "subscription_instance_relations",
        "subscription_instances",
        ["parent_id"],
        ["subscription_instance_id"],
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        "subscription_instance_relations_child_id_fkey",
        "subscription_instance_relations",
        "subscription_instances",
        ["child_id"],
        ["subscription_instance_id"],
        ondelete="CASCADE",
    )



import time
import uuid
from http import HTTPStatus
from threading import Condition, Event
from unittest import mock
from uuid import uuid4

import pytest
from sqlalchemy import select

from oauth2_lib.fastapi import OIDCUserModel
from orchestrator.config.assignee import Assignee
from orchestrator.db import (
    ProcessStepTable,
    ProcessSubscriptionTable,
    ProcessTable,
    SubscriptionTable,
    db,
)
from orchestrator.security import authenticate
from orchestrator.services.processes import shutdown_thread_pool
from orchestrator.services.settings import get_engine_settings
from orchestrator.settings import app_settings
from orchestrator.workflow import ProcessStatus, done, init, step, workflow
from test.unit_tests.helpers import URL_STR_TYPE
from test.unit_tests.workflows import WorkflowInstanceForTests

test_condition = Condition()


@pytest.fixture
def long_running_workflow():
    @step("Long Running Step")
    def long_running_step():
        with test_condition:
            test_condition.wait()
        return {"done": True}

    @workflow("Long Running Workflow")
    def long_running_workflow_py():
        return init >> long_running_step >> long_running_step >> done

    with WorkflowInstanceForTests(long_running_workflow_py, "long_running_workflow_py"):
        yield "long_running_workflow_py"


@pytest.fixture
def started_process(test_workflow, generic_subscription_1):
    process_id = uuid4()
    process = ProcessTable(
        process_id=process_id, workflow_id=test_workflow.workflow_id, last_status=ProcessStatus.SUSPENDED
    )
    init_step = ProcessStepTable(process_id=process_id, name="Start", status="success", state={})
    insert_step = ProcessStepTable(
        process_id=process_id,
        name="Insert UUID in state",
        status="success",
        state={"subscription_id": generic_subscription_1},
    )
    check_step = ProcessStepTable(
        process_id=process_id,
        name="Test that it is a string now",
        status="success",
        state={"subscription_id": generic_subscription_1},
    )
    step = ProcessStepTable(
        process_id=process_id, name="Modify", status="suspend", state={"subscription_id": generic_subscription_1}
    )

    process_subscription = ProcessSubscriptionTable(process_id=process_id, subscription_id=generic_subscription_1)

    db.session.add(process)
    db.session.add(init_step)
    db.session.add(insert_step)
    db.session.add(check_step)
    db.session.add(step)
    db.session.add(process_subscription)
    db.session.commit()

    return process_id


def test_show(test_client, started_process):
    response = test_client.get(f"/api/processes/{started_process}")
    assert HTTPStatus.OK == response.status_code
    process = response.json()
    assert "workflow_for_testing_processes_py", process["workflow_name"]


def test_show_invalid_uuid(test_client):
    response = test_client.get("/api/processes/wrong")
    assert response.status_code == 422


def test_show_not_found(test_client, started_process):
    response = test_client.get("/api/processes/120d31e4-9166-47cb-ad37-b78072c7ab8b")
    assert HTTPStatus.NOT_FOUND == response.status_code


def test_delete_process(responses, test_client, started_process):
    processes = test_client.get("/api/processes").json()
    before_delete_count = len(processes)

    response = test_client.delete(f"/api/processes/{started_process}")
    assert HTTPStatus.NO_CONTENT == response.status_code
    assert before_delete_count - 1 == len(test_client.get("/api/processes").json())


def test_delete_process_404(test_client, started_process):
    response = test_client.delete(f"/api/processes/{uuid4()}")
    assert HTTPStatus.NOT_FOUND == response.status_code


def test_long_running_pause(test_client, long_running_workflow):
    app_settings.TESTING = False
    # Start the workflow
    response = test_client.post(f"/api/processes/{long_running_workflow}", json=[{}])
    assert (
        HTTPStatus.CREATED == response.status_code
    ), f"Invalid response status code (response data: {response.json()})"

    process_id = response.json()["id"]

    response = test_client.get(f"api/processes/{process_id}")
    assert HTTPStatus.OK == response.status_code

    # Let it run until the first lock step is started
    time.sleep(1)

    response = test_client.put("/api/settings/status", json={"global_lock": True})
    assert response.json()["global_lock"] is True
    assert response.json()["running_processes"] == 1
    assert response.json()["global_status"] == "PAUSING"

    # Let it run until completing the first lock step
    with test_condition:
        test_condition.notify_all()
    time.sleep(1)

    # Make sure the running_processes is decreased and status is updated.
    response = test_client.get("/api/settings/status")
    assert response.json()["global_lock"] is True
    assert response.json()["running_processes"] == 0
    assert response.json()["global_status"] == "PAUSED"

    response = test_client.get(f"api/processes/{process_id}")
    assert len(response.json()["steps"]) == 4
    assert response.json()["current_state"]["done"] is True
    # assume ordered steplist
    assert response.json()["steps"][3]["status"] == "pending"

    response = test_client.put("/api/settings/status", json={"global_lock": False})

    # Make sure it started again
    time.sleep(1)

    assert response.json()["global_lock"] is False
    assert response.json()["running_processes"] == 1
    assert response.json()["global_status"] == "RUNNING"

    # Let it finish after second lock step
    with test_condition:
        test_condition.notify_all()
    time.sleep(1)

    response = test_client.get(f"api/processes/{process_id}")
    assert HTTPStatus.OK == response.status_code
    # assume ordered steplist
    assert response.json()["steps"][3]["status"] == "complete"

    app_settings.TESTING = True


def test_service_unavailable_engine_locked(test_client, test_workflow):
    engine_settings = get_engine_settings()
    engine_settings.global_lock = True
    db.session.flush()

    response = test_client.post(f"/api/processes/{test_workflow}", json=[{}])

    assert HTTPStatus.SERVICE_UNAVAILABLE == response.status_code


def test_complete_workflow(test_client, test_workflow):
    response = test_client.post(f"/api/processes/{test_workflow.name}", json=[{}])

    assert (
        HTTPStatus.CREATED == response.status_code
    ), f"Invalid response status code (response data: {response.json()})"

    process_id = response.json()["id"]

    response = test_client.get(f"api/processes/{process_id}")
    assert HTTPStatus.OK == response.status_code

    process = response.json()
    assert Assignee.CHANGES == process["assignee"]
    assert "suspended" == process["last_status"]

    steps = process["steps"]
    assert "success" == steps[0]["status"]

    response = test_client.get(f"/api/processes/{process_id}")
    assert response.json()["form"] == {
        "title": "unknown",
        "type": "object",
        "properties": {"generic_select": {"$ref": "#/$defs/TestChoice"}},
        "additionalProperties": False,
        "required": ["generic_select"],
        "$defs": {"TestChoice": {"enum": ["A", "B", "C"], "title": "TestChoice", "type": "string"}},
    }

    # Check type validation
    user_input = {"generic_select": 123}
    response = test_client.put(f"/api/processes/{process_id}/resume", json=[user_input])
    assert HTTPStatus.BAD_REQUEST == response.status_code
    assert response.json()["validation_errors"] == [
        {
            "ctx": {"expected": "'A', 'B' or 'C'"},
            "input": 123,
            "loc": ["generic_select"],
            "msg": "Input should be 'A', 'B' or 'C'",
            "type": "enum",
        }
        | URL_STR_TYPE
    ]

    # Check choice validation
    user_input = {"generic_select": "123"}
    response = test_client.put(f"/api/processes/{process_id}/resume", json=[user_input])
    assert HTTPStatus.BAD_REQUEST == response.status_code
    assert response.json()["validation_errors"] == [
        {
            "ctx": {"expected": "'A', 'B' or 'C'"},
            "input": "123",
            "loc": ["generic_select"],
            "msg": "Input should be 'A', 'B' or 'C'",
            "type": "enum",
        }
        | URL_STR_TYPE
    ]

    response = test_client.get(f"api/processes/{process_id}")

    process = response.json()
    assert "suspended" == process["last_status"]

    # Now for real
    user_input = {"generic_select": "A"}

    response = test_client.put(f"/api/processes/{process_id}/resume", json=[user_input])
    assert HTTPStatus.NO_CONTENT == response.status_code

    process = test_client.get(f"api/processes/{process_id}").json()
    assert "completed" == process["last_status"]


def test_abort_process(test_client, started_process):
    response = test_client.put(f"/api/processes/{started_process}/abort")
    assert HTTPStatus.NO_CONTENT == response.status_code

    aborted_process = test_client.get(f"/api/processes/{started_process}").json()
    assert "aborted" == aborted_process["last_status"]


def test_new_process_invalid_body(test_client, long_running_workflow):
    response = test_client.post(f"/api/processes/{long_running_workflow}", json=[{"wrong": "body"}])
    assert HTTPStatus.BAD_REQUEST == response.status_code


def test_new_process_invalid_Content_Type(test_client):
    response = test_client.post("/api/processes/terminate_sn8_light_path")
    assert HTTPStatus.UNPROCESSABLE_ENTITY == response.status_code


def test_new_process_post_inconsistent_data(test_client):
    response = test_client.post("/api/processes/terminate_sn8_light_path", json={})
    assert HTTPStatus.UNPROCESSABLE_ENTITY == response.status_code


def test_404_resume(test_client):
    response = test_client.put(f"/api/processes/{uuid.uuid4()}/resume", json={})
    assert HTTPStatus.NOT_FOUND == response.status_code


def test_resume_validations(test_client, started_process):
    process_info_before = test_client.get(f"/api/processes/{started_process}").json()

    response = test_client.put(f"/api/processes/{started_process}/resume", json=[{"generic_select": 123}])
    assert HTTPStatus.BAD_REQUEST == response.status_code
    assert [
        {
            "ctx": {"expected": "'A', 'B' or 'C'"},
            "input": 123,
            "loc": ["generic_select"],
            "msg": "Input should be 'A', 'B' or 'C'",
            "type": "enum",
        }
        | URL_STR_TYPE
    ] == response.json()["validation_errors"]
    process_info_after = test_client.get(f"/api/processes/{started_process}").json()
    excuted_steps_before = [step for step in process_info_before["steps"] if step.get("executed")]
    excuted_steps_after = [step for step in process_info_after["steps"] if step.get("executed")]
    assert len(excuted_steps_after) == len(excuted_steps_before)
    assert process_info_after["last_status"] == "suspended"


def test_resume_with_empty_form(test_client, started_process):
    # Set a default value for the only input so we can submit an empty form
    step = db.session.scalars(
        select(ProcessStepTable).filter(
            ProcessStepTable.name == "Modify", ProcessStepTable.process_id == started_process
        )
    ).one()
    step.state["generic_select"] = "A"
    db.session.add(step)
    db.session.flush()

    process_info_before = test_client.get(f"/api/processes/{started_process}").json()
    response = test_client.put(f"/api/processes/{started_process}/resume", json=[{"generic_select": "A"}])

    assert HTTPStatus.NO_CONTENT == response.status_code
    process_info_after = test_client.get(f"/api/processes/{started_process}").json()
    excuted_steps_before = [step for step in process_info_before["steps"] if step.get("executed")]
    excuted_steps_after = [step for step in process_info_after["steps"] if step.get("executed")]
    assert len(excuted_steps_after) > len(excuted_steps_before)
    assert process_info_after["last_status"] == "completed"


def test_resume_happy_flow(test_client, started_process):
    process_info_before = test_client.get(f"/api/processes/{started_process}").json()
    response = test_client.put(f"/api/processes/{started_process}/resume", json=[{"generic_select": "A"}])

    assert HTTPStatus.NO_CONTENT == response.status_code
    process_info_after = test_client.get(f"/api/processes/{started_process}").json()
    excuted_steps_before = [step for step in process_info_before["steps"] if step.get("executed")]
    excuted_steps_after = [step for step in process_info_after["steps"] if step.get("executed")]
    assert len(excuted_steps_after) > len(excuted_steps_before)
    assert process_info_after["last_status"] == "completed"


def test_resume_with_incorrect_workflow_status(test_client, started_process):
    process = db.session.get(ProcessTable, started_process)
    # setup DB so it looks like this workflow is already resumed
    process.last_status = ProcessStatus.RUNNING
    process.failed_reason = ""
    db.session.commit()

    process_info_before = test_client.get(f"/api/processes/{started_process}").json()
    response = test_client.put(f"/api/processes/{started_process}/resume", json=[{"generic_select": "A"}])
    assert 409 == response.status_code
    process_info_after = test_client.get(f"/api/processes/{started_process}").json()
    excuted_steps_before = [step for step in process_info_before["steps"] if step.get("executed")]
    excuted_steps_after = [step for step in process_info_after["steps"] if step.get("executed")]
    assert len(excuted_steps_after) == len(excuted_steps_before)
    assert process_info_after["last_status"] == "running"


def test_try_resume_completed_workflow(test_client, started_process):
    process = db.session.get(ProcessTable, started_process)
    assert process
    # setup DB so it looks like this workflow is already completed
    process.last_status = ProcessStatus.COMPLETED
    process.failed_reason = ""
    db.session.commit()

    response = test_client.put(f"/api/processes/{started_process}/resume", json=[{}])
    assert 409 == response.status_code


def test_try_resume_resumed_workflow(test_client, started_process):
    process = db.session.get(ProcessTable, started_process)
    assert process
    # setup DB so it looks like this workflow has already been resumed
    process.last_status = ProcessStatus.RESUMED
    process.failed_reason = ""
    db.session.add(process)
    db.session.commit()

    response = test_client.put(f"/api/processes/{started_process}/resume", json=[{}])
    assert 409 == response.status_code


def test_processes_filterable(test_client, mocked_processes, generic_subscription_2, generic_subscription_1):
    response = test_client.get("/api/processes/")
    assert HTTPStatus.OK == response.status_code
    processes = response.json()

    assert len(processes) == 9
    assert "workflow_for_testing_processes_py", processes[0]["workflow"]

    response = test_client.get("/api/processes?filter=lastStatus,completed")
    assert len(response.json()) == 3
    response = test_client.get("/api/processes?filter=lastStatus,suspended")
    assert len(response.json()) == 2
    response = test_client.get("/api/processes?filter=lastStatus,resumed")
    assert len(response.json()) == 2

    product_name = db.session.get(SubscriptionTable, generic_subscription_1).product.name
    response = test_client.get(f"/api/processes?filter=product,{product_name}")
    assert len(response.json()) == 4
    response = test_client.get("/api/processes?sort=assignee,asc")
    assert response.json()[0]["assignee"] == "NOC"
    response = test_client.get("/api/processes?sort=startedAt&filter=isTask,true")
    assert len(response.json()) == 4


def test_processes_filterable_response_model(
    test_client, mocked_processes, generic_subscription_2, generic_subscription_1
):
    response = test_client.get("/api/processes/?sort=startedAt,asc")
    assert HTTPStatus.OK == response.status_code
    processes = response.json()
    assert len(processes) == 9
    process = processes[0]

    assert len(process["subscriptions"]) == 1

    # Check if the other fields are filled with correct data
    del process["process_id"]  # skip process_id as it's dynamic
    del process["product_id"]  # skip product_id as it's dynamic
    del process["workflow_id"]  # skip workflow_id as it's dynamic
    del process["subscriptions"]
    assert process == {
        "customer_id": "2f47f65a-0911-e511-80d0-005056956c1a",
        "workflow_name": "workflow_for_testing_processes_py",
        "is_task": False,
        "assignee": "SYSTEM",
        "last_status": "completed",
        "last_step": "Modify",
        "failed_reason": None,
        "traceback": None,
        "created_by": None,
        "started_at": 1578994200.0,
        "last_modified_at": 1578994800.0,
        "current_state": None,
        "steps": None,
        "form": None,
        "workflow_target": "CREATE",
    }


def test_processes_filterable_response_model_contains_product_info(
    test_client, mocked_processes, generic_subscription_2, generic_subscription_1
):
    response = test_client.get("/api/processes/?sort=startedAt,asc")
    assert HTTPStatus.OK == response.status_code
    processes = response.json()
    assert len(processes) == 9
    process = processes[0]

    assert len(process["subscriptions"]) == 1
    assert process["subscriptions"][0]["product"]["tag"] == "GEN1"
    assert process["subscriptions"][0]["product"]["name"] == "Product 1"


def test_resume_all_processes(test_client, mocked_processes_resumeall):
    """Test resuming all processes."""
    response = test_client.put("/api/processes/resume-all")
    assert HTTPStatus.OK == response.status_code
    assert response.json()["count"] == 3


def test_resume_all_processes_multiple_calls(test_client, mocked_processes_resumeall):
    """Test only 1 of multiple resume-all calls is successful.

    This uses the "MemoryDistlockManager" reference implementation.
    """
    event = Event()

    def resume_noop(*args, **kwargs):
        event.wait(2)  # To keep the lock open for a while

    # Disable Testing setting since we want to run async
    app_settings.TESTING = False

    with mock.patch("orchestrator.services.processes.resume_process", new=resume_noop):
        responses = [test_client.put("/api/processes/resume-all") for _ in range(5)]
        responses.sort(key=lambda r: r.status_code)
        event.set()

    app_settings.TESTING = True

    assert responses[0].status_code == HTTPStatus.OK
    assert responses[0].json()["count"] == 3
    assert all(response.status_code == HTTPStatus.CONFLICT for response in responses[1:])

    # Wait for async tasks to finish to prevent DB session conflicts
    shutdown_thread_pool()


def test_resume_all_processes_nothing_to_do(test_client):
    """Test resuming all process when there are no process to be resumed."""
    response = test_client.put("/api/processes/resume-all")
    assert HTTPStatus.OK == response.status_code
    assert response.json()["count"] == 0


def test_resume_all_processes_value_error(test_client, mocked_processes_resumeall, caplog):
    """Test resuming all processes where one raises ValueError."""
    with mock.patch("orchestrator.services.processes.resume_process") as mocked_resume:
        mocked_resume.side_effect = [None, ValueError("This workflow cannot be resumed"), None]
        response = test_client.put("/api/processes/resume-all")
    assert HTTPStatus.OK == response.status_code
    assert response.json()["count"] == 3  # returns 3 because it's async
    assert "Failed to resume process" in caplog.text  # log should confirm 1 process was not resumed
    assert "Completed resuming processes" in caplog.text


@pytest.mark.parametrize(
    "oidc_user, reporter, expected_user",
    [
        (OIDCUserModel({"name": "alice"}), None, "alice"),
        (OIDCUserModel({"user_name": "alice"}), None, ""),  # without overriding OIDCUserModel, user_name has no effect
        (OIDCUserModel({"name": "alice"}), "bob", "bob"),  # reporter param has precedence over oidc user
    ],
)
def test_create_process_reporter(test_client, fastapi_app, oidc_user, reporter, expected_user):
    # given
    url_params = {"reporter": reporter} if reporter is not None else {}
    fastapi_depends = {authenticate: lambda: oidc_user}
    with (
        mock.patch("orchestrator.api.api_v1.endpoints.processes.start_process") as mock_start_process,
        mock.patch.dict(fastapi_app.dependency_overrides, fastapi_depends),
    ):
        mock_start_process.return_value = uuid.uuid4()
        # when
        response = test_client.post("/api/processes/fake_workflow", json=[], params=url_params)

    # then
    assert response.status_code == 201, response.text
    mock_start_process.assert_called()
    assert mock_start_process.mock_calls[0].kwargs["user"] == expected_user



from more_itertools.more import one
from more_itertools.recipes import first_true
from sqlalchemy import not_, select
from sqlalchemy.orm import joinedload

from orchestrator.db import ProductTable, WorkflowTable, db
from orchestrator.db.models import FixedInputTable
from orchestrator.services import products
from orchestrator.services.products import get_products
from orchestrator.targets import Target
from orchestrator.utils.fixed_inputs import fixed_input_configuration as fi_configuration
from orchestrator.workflows import ALL_WORKFLOWS


def test_all_workflows_have_matching_targets_and_descriptions():
    for key, lazy_wf in ALL_WORKFLOWS.items():
        workflow = lazy_wf.instantiate()
        db_workflow = db.session.scalars(WorkflowTable.select().where(WorkflowTable.name == key)).first()
        if db_workflow:
            # Test workflows might not exist in the database
            assert workflow.target == db_workflow.target
            assert workflow.name == db_workflow.name
            assert workflow.description == db_workflow.description


def test_all_products_have_at_least_one_workflow():
    prods_without_wf = db.session.scalars(
        select(ProductTable).where(not_(ProductTable.workflows.any())).with_only_columns(ProductTable.name)
    ).all()
    assert len(prods_without_wf) == 0, (
        f"These products do not have a workflow " f"associated with them: {', '.join(prods_without_wf)}."
    )


def test_all_non_system_workflows_have_at_least_one_product(generic_product_1):
    wfs_without_prod = db.session.scalars(
        WorkflowTable.select()
        .where(WorkflowTable.target != Target.SYSTEM, not_(WorkflowTable.products.any()))
        .with_only_columns(WorkflowTable.name)
    ).all()

    assert len(wfs_without_prod) == 0, (
        f"These workflows do not have a product " f"associated with them: {', '.join(wfs_without_prod)}."
    )


def test_all_active_products_have_a_modify_note():
    """Note: when fail: create a migration for the new product with `helpers/create_missing_modify_note_workflows()`."""
    modify_workflow = db.session.scalars(WorkflowTable.select().where(WorkflowTable.name == "modify_note")).first()

    products = get_products(filters=[ProductTable.status == "active"])
    result = [product.name for product in products if modify_workflow not in product.workflows]
    assert not len(result), f"These products do not have a modify_note workflow {', '.join(result)}."


def test_db_fixed_input_config():
    fixed_input_configuration = fi_configuration()

    product_tags = products.get_tags()
    fixed_inputs = db.session.scalars(select(FixedInputTable).options(joinedload(FixedInputTable.product))).all()

    data = {"fixed_inputs": [], "by_tag": {}}
    for tag in product_tags:
        data["by_tag"][tag] = []

    for fi in fixed_inputs:
        fi_data = first_true(
            fixed_input_configuration["fixed_inputs"], None, lambda i: i["name"] == fi.name  # noqa: B023
        )
        assert fi_data, fi
        assert fi.value in fi_data["values"], fi

        tag_data = {one(fi) for fi in fixed_input_configuration["by_tag"][fi.product.tag]}
        tag_data_required = {one(fi) for fi in fixed_input_configuration["by_tag"][fi.product.tag] if fi[one(fi)]}
        assert tag_data, fi
        assert not {fi.name for fi in fi.product.fixed_inputs} - set(tag_data), fi.product.name
        assert not set(tag_data_required) - {fi.name for fi in fi.product.fixed_inputs}, fi.product.name



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from uuid import UUID

from fastapi.routing import APIRouter
from pytz import timezone
from sqlalchemy import select

from orchestrator.api.models import delete
from orchestrator.db import SubscriptionCustomerDescriptionTable, db
from orchestrator.utils.redis import delete_subscription_from_redis
from orchestrator.websocket import invalidate_subscription_cache

router = APIRouter()


def get_customer_description_by_customer_subscription(
    customer_id: str, subscription_id: UUID
) -> SubscriptionCustomerDescriptionTable | None:
    stmt = select(SubscriptionCustomerDescriptionTable).filter(
        SubscriptionCustomerDescriptionTable.customer_id == customer_id,
        SubscriptionCustomerDescriptionTable.subscription_id == str(subscription_id),
    )
    return db.session.scalars(stmt).one_or_none()


@delete_subscription_from_redis()
async def create_subscription_customer_description(
    customer_id: str, subscription_id: UUID, description: str
) -> SubscriptionCustomerDescriptionTable:
    customer_description = SubscriptionCustomerDescriptionTable(
        customer_id=customer_id,
        subscription_id=subscription_id,
        description=description,
    )
    db.session.add(customer_description)
    db.session.commit()
    await invalidate_subscription_cache(customer_description.subscription_id)
    return customer_description


@delete_subscription_from_redis()
async def update_subscription_customer_description(
    customer_description: SubscriptionCustomerDescriptionTable, description: str, created_at: datetime | None = None
) -> SubscriptionCustomerDescriptionTable:
    customer_description.description = description
    customer_description.created_at = created_at if created_at else datetime.now(tz=timezone("UTC"))
    db.session.commit()
    await invalidate_subscription_cache(customer_description.subscription_id)
    return customer_description


@delete_subscription_from_redis()
async def delete_subscription_customer_description_by_customer_subscription(
    customer_id: str, subscription_id: UUID
) -> SubscriptionCustomerDescriptionTable | None:
    customer_description = get_customer_description_by_customer_subscription(customer_id, subscription_id)
    if not customer_description:
        return None

    delete(SubscriptionCustomerDescriptionTable, customer_description.id)
    await invalidate_subscription_cache(customer_description.subscription_id)
    return customer_description



import pytest
from orchestrator.db import ProductTable

from products.product_types.example4 import Example4
from test.unit_tests.workflows import assert_complete, extract_state, run_workflow


@pytest.mark.workflow
def test_happy_flow(responses):
    # given

    # TODO insert additional mocks, if needed (ImsMocks)

    product = db.session.scalars(select(ProductTable).where(ProductTable.name == "example4")).one()

    # when

    init_state = {
        "customer_id": customer_id,
        # TODO add initial state
    }

    result, process, step_log = run_workflow("create_example4", [{"product": product.product_id}, init_state])

    # then

    assert_complete(result)
    state = extract_state(result)

    subscription = Example4.from_subscription(state["subscription_id"])
    assert subscription.status == "active"
    assert subscription.description == "TODO add correct description"



from products.product_blocks.example1 import AnnotatedInt, ExampleStrEnum1


def must_be_unused_to_change_mode_validator(example_str_enum_1: ExampleStrEnum1) -> ExampleStrEnum1:
    if False:  # TODO: implement validation for example_str_enum_1
        raise ValueError("Mode can only be changed when there are no services attached to it")

    return example_str_enum_1


def annotated_int_must_be_unique_validator(annotated_int: AnnotatedInt) -> AnnotatedInt:
    if False:  # TODO: implement validation for annotated_int
        raise ValueError("annotated_int must be unique for example1")

    return annotated_int



from orchestrator.domain.base import SubscriptionModel
from orchestrator.types import SubscriptionLifecycle

from products.product_blocks.example2 import Example2Block, Example2BlockInactive, Example2BlockProvisioning


class Example2Inactive(SubscriptionModel, is_base=True):
    example2: Example2BlockInactive


class Example2Provisioning(Example2Inactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    example2: Example2BlockProvisioning


class Example2(Example2Provisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    example2: Example2Block



import pytest

from orchestrator.db import ProductBlockTable, db
from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle


@pytest.fixture
def test_product_sub_block_two():
    class SubBlockTwoForTestInactive(ProductBlockModel, product_block_name="SubBlockTwoForTest"):
        int_field_2: int  # TODO #430 inactive productblocks should not have required fields

    class SubBlockTwoForTestProvisioning(SubBlockTwoForTestInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
        int_field_2: int

    class SubBlockTwoForTest(SubBlockTwoForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        int_field_2: int

    return SubBlockTwoForTestInactive, SubBlockTwoForTestProvisioning, SubBlockTwoForTest


@pytest.fixture
def test_product_sub_block_two_db(resource_type_int_2):
    sub_block = ProductBlockTable(
        name="SubBlockTwoForTest", description="Test Sub Block Two", tag="TEST", status="active"
    )

    sub_block.resource_types = [resource_type_int_2]

    db.session.add(sub_block)
    db.session.commit()
    return sub_block



from collections.abc import Callable

from sqlalchemy import Column, Select
from sqlalchemy.inspection import inspect
from sqlalchemy.sql import expression

from orchestrator.db import ProcessSubscriptionTable, ProcessTable, ProductTable, SubscriptionTable, WorkflowTable
from orchestrator.db.filters import create_memoized_field_list
from orchestrator.db.sorting import QueryType, SortOrder, generic_column_sort, generic_sort
from orchestrator.utils.helpers import to_camel

PROCESS_CAMEL_SORT = {
    to_camel(key): generic_column_sort(value, ProcessTable) for key, value in inspect(ProcessTable).columns.items()
}
PROCESS_SNAKE_SORT = {
    key: generic_column_sort(value, ProcessTable) for key, value in inspect(ProcessTable).columns.items()
}


def generic_process_relation_sort(
    field: Column, join_tables: list | None = None
) -> Callable[[QueryType, SortOrder], QueryType]:
    join_tables = join_tables or []

    def sort_function(query: QueryType, order: SortOrder) -> QueryType:
        # SQLAlchemy can't resolve the query correctly when we join the same table multiple times. A refactor for the sorting logic that
        # ensures each table is joined once for the sort order is required. For now, we access an internal member `_setup_join` to check whether a table
        # has already been joined
        if isinstance(query, Select):
            joined_tables = [str(t[0]) for t in query._setup_joins]

            for table in join_tables:
                if (tablename := getattr(table, "__tablename__", "")) and tablename not in joined_tables:
                    query = query.join(table)

        order_by_expr = expression.desc if order == SortOrder.DESC else expression.asc
        return query.order_by(order_by_expr(field))

    return sort_function


PROCESS_PRODUCT_SORT = {
    to_camel(key if "product" in key else f"product_{key}"): generic_process_relation_sort(
        value, [ProcessSubscriptionTable, SubscriptionTable, ProductTable]
    )
    for key, value in inspect(ProductTable).columns.items()
}


PROCESS_SORT_FUNCTIONS_BY_COLUMN = (
    PROCESS_PRODUCT_SORT
    | PROCESS_CAMEL_SORT
    | PROCESS_SNAKE_SORT
    | {
        "workflowName": generic_process_relation_sort(WorkflowTable.name, [WorkflowTable]),
        "workflowTarget": generic_process_relation_sort(WorkflowTable.target, [WorkflowTable]),
        "subscriptions": generic_process_relation_sort(
            SubscriptionTable.description, [ProcessSubscriptionTable, SubscriptionTable]
        ),
    }
)
process_sort_fields = create_memoized_field_list(PROCESS_SORT_FUNCTIONS_BY_COLUMN)
sort_processes = generic_sort(PROCESS_SORT_FUNCTIONS_BY_COLUMN)



import asyncio
import contextlib
from collections import namedtuple
from unittest import mock

import pytest
from fastapi import WebSocket

from orchestrator.settings import app_settings
from orchestrator.websocket import WS_CHANNELS
from orchestrator.websocket.websocket_manager import WebSocketManager


@pytest.fixture
def redis_uri() -> str:
    return str(app_settings.CACHE_URI)


@pytest.fixture
async def websocket_managers(redis_uri):
    async def create_manager():
        wsm = WebSocketManager(True, redis_uri)
        await wsm.connect_redis()
        return wsm

    managers = [await create_manager() for _ in range(2)]

    yield managers

    for manager in managers:
        await manager.disconnect_redis()


FakeWsClient = namedtuple("FakeWsClient", ["client", "stop_event"])


@pytest.fixture
async def websocket_clients(websocket_managers, test_client):
    wsm1, wsm2 = websocket_managers

    def make_client(name: str) -> FakeWsClient:
        # Create a fake websocket client based the WebSocket class
        mock_client = mock.MagicMock(spec_set=WebSocket, name=name)
        stop_event = asyncio.Event()

        block_seconds = 3

        async def blocking_receive_text():
            # Block for block_seconds or until the stop_event has been set
            with contextlib.suppress(asyncio.TimeoutError):
                # Suppress timeout error, if something's wrong this will be seen in the main testcase
                await asyncio.wait_for(stop_event.wait(), block_seconds)
            raise StopAsyncIteration  # End the

        # Make the BroadcastWebsocketManager.receiver() function block, and thereby the .connect() loop
        mock_client.receive_text.side_effect = blocking_receive_text

        def mock_send_text(*_args):
            stop_event.set()

        # When there is a call to WebSocket.send_text(), unblock .receiver() to close the loop
        mock_client.send_text.side_effect = mock_send_text
        return FakeWsClient(mock_client, stop_event)

    # Prepare 2 fake clients
    clients: list[FakeWsClient] = [make_client("Fake Websocket Client 1"), make_client("Fake Websocket Client 2")]

    async with asyncio.TaskGroup() as taskgroup:
        # 'Connect' client 1 to WSM 1
        taskgroup.create_task(wsm1.connect(clients[0].client, WS_CHANNELS.EVENTS))
        # 'Connect' client 2 to WSM 2
        taskgroup.create_task(wsm2.connect(clients[1].client, WS_CHANNELS.EVENTS))

        # Continue the testcase - tasks are cleared on teardown
        yield clients


async def test_broadcast(websocket_managers: list[WebSocketManager], websocket_clients: list[FakeWsClient]):
    """Integration test for websocket broadcasting, requires redis."""
    wsm1, wsm2 = websocket_managers

    # given: 2 websocket clients, each connected to a different WSM
    assert len(websocket_clients) == 2, "Fixture should return 2 clients"
    assert len(wsm1._backend.connected) == 1  # type: ignore
    assert len(wsm2._backend.connected) == 1  # type: ignore

    # when: we broadcast something on WSM 2
    await wsm2.broadcast_data([WS_CHANNELS.EVENTS], {"message": "foobar"})

    # then: both clients should have been sent the broadcasted message
    for client, stop_event in websocket_clients:
        await asyncio.wait_for(stop_event.wait(), 1)
        assert mock.call('{"message":"foobar"}') in client.send_text.mock_calls



from enum import IntEnum

from orchestrator.domain.base import SubscriptionModel
from orchestrator.types import SubscriptionLifecycle

from products.product_blocks.example1 import Example1Block, Example1BlockInactive, Example1BlockProvisioning


class FixedInput1(IntEnum):
    _1 = 1
    _10 = 10
    _100 = 100
    _1000 = 1000


class Example1Inactive(SubscriptionModel, is_base=True):
    fixed_input_1: FixedInput1
    example1: Example1BlockInactive


class Example1Provisioning(Example1Inactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    fixed_input_1: FixedInput1
    example1: Example1BlockProvisioning


class Example1(Example1Provisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    fixed_input_1: FixedInput1
    example1: Example1Block



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from http import HTTPStatus
from unittest import mock

from sqlalchemy.exc import OperationalError


def test_get_health(test_client):
    response = test_client.get("/api/health/")
    assert HTTPStatus.OK == response.status_code
    assert response.json() == "OK"


@mock.patch("orchestrator.db.db.session")
def test_get_health_no_connection(mock_session, test_client):
    mock_session.execute.side_effect = OperationalError("THIS", "IS", "KABOOM")
    response = test_client.get("/api/health/")
    assert response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR



from collections.abc import Iterable

from orchestrator.types import filter_nonetype, get_origin_and_args, is_union_type


def _to_product_block_field_type_iterable(product_block_field_type: type | tuple[type]) -> Iterable[type]:
    """Return an iterable of types in the given product block field.

    Notes:
        There is a type-checking pattern (if optional, if union, if list/tuple, etc) which occurs multiple times in
        base.py and other parts of the codebase. We should try to combine those into one function.
        However, this function works based on field types as returned by
        DomainModel._get_depends_on_product_block_types() so it is not directly applicable.
    """
    _origin, args = get_origin_and_args(product_block_field_type)

    if is_union_type(product_block_field_type):
        return list(filter_nonetype(args))

    if isinstance(product_block_field_type, tuple):
        return product_block_field_type

    return [product_block_field_type]



# Copyright 2022-2024 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from uuid import UUID

import strawberry
import structlog

from oauth2_lib.strawberry import authenticated_mutation_field
from orchestrator.db.models import SubscriptionCustomerDescriptionTable
from orchestrator.domain.customer_description import (
    create_subscription_customer_description,
    delete_subscription_customer_description_by_customer_subscription,
    get_customer_description_by_customer_subscription,
    update_subscription_customer_description,
)
from orchestrator.graphql.schemas.customer_description import CustomerDescription
from orchestrator.graphql.types import MutationError, NotFoundError

logger = structlog.get_logger(__name__)


async def upsert_customer_description(
    customer_id: str, subscription_id: UUID, description: str
) -> SubscriptionCustomerDescriptionTable | NotFoundError:
    current_description = get_customer_description_by_customer_subscription(customer_id, subscription_id)

    if current_description:
        return await update_subscription_customer_description(current_description, description)
    return await create_subscription_customer_description(customer_id, subscription_id, description)


async def resolve_upsert_customer_description(
    customer_id: str, subscription_id: UUID, description: str
) -> CustomerDescription | NotFoundError | MutationError:
    try:
        customer_description = await upsert_customer_description(customer_id, subscription_id, description)
    except Exception:
        return NotFoundError(message="Subscription not found")
    return CustomerDescription.from_pydantic(customer_description)  # type: ignore


async def resolve_remove_customer_description(
    customer_id: str, subscription_id: UUID
) -> CustomerDescription | NotFoundError | MutationError:
    description = await delete_subscription_customer_description_by_customer_subscription(
        customer_id=customer_id, subscription_id=subscription_id
    )
    if not description:
        return NotFoundError(message="Customer description not found")
    return CustomerDescription.from_pydantic(description)


@strawberry.type(description="Customer subscription description mutations")
class CustomerSubscriptionDescriptionMutation:
    upsert_customer_description = authenticated_mutation_field(
        resolver=resolve_upsert_customer_description,
        description="Create or update customer description",
    )
    remove_customer_description = authenticated_mutation_field(
        resolver=resolve_remove_customer_description,
        description="Delete customer description",
    )



from os import getenv
from unittest import mock
from unittest.mock import Mock

import pytest

from orchestrator import app_settings
from orchestrator.domain.base import SubscriptionModel
from orchestrator.services.subscriptions import build_extended_domain_model
from orchestrator.utils.get_subscription_dict import get_subscription_dict
from orchestrator.utils.redis import to_redis


@mock.patch.object(app_settings, "CACHE_DOMAIN_MODELS", False)
@mock.patch("orchestrator.utils.get_subscription_dict._generate_etag")
async def test_get_subscription_dict_db(generate_etag, generic_subscription_1):
    generate_etag.side_effect = Mock(return_value="etag-mock")
    await get_subscription_dict(generic_subscription_1)
    assert generate_etag.called


@pytest.mark.skipif(
    not getenv("AIOCACHE_DISABLE", "0") == "0", reason="AIOCACHE must be enabled for this test to do anything"
)
@mock.patch("orchestrator.utils.get_subscription_dict._generate_etag")
async def test_get_subscription_dict_cache(generate_etag, generic_subscription_1, cache_fixture):
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    extended_model = build_extended_domain_model(subscription)

    # Add domainmodel to cache
    to_redis(extended_model)
    cache_fixture.extend(
        [f"orchestrator:domain:{generic_subscription_1}", f"orchestrator:domain:etag:{generic_subscription_1}"]
    )

    generate_etag.side_effect = Mock(return_value="etag-mock")
    await get_subscription_dict(generic_subscription_1)
    assert not generate_etag.called



from unittest import mock
from uuid import uuid4

import pytest

from orchestrator.config.assignee import Assignee
from orchestrator.db import ProcessStepTable, ProcessTable, WorkflowTable, db
from orchestrator.targets import Target
from orchestrator.workflow import ProcessStatus
from test.unit_tests.workflows import assert_complete, assert_state, extract_state, run_workflow


@pytest.fixture
def waiting_process():
    state = {"foo": "bar"}
    pid = uuid4()
    workflow_id = uuid4()
    waiting_workflow = WorkflowTable(
        workflow_id=workflow_id, name="Waiting workflow", target=Target.SYSTEM, description="Description"
    )

    process = ProcessTable(
        process_id=pid,
        workflow_id=workflow_id,
        last_status=ProcessStatus.WAITING,
        assignee=Assignee.SYSTEM,
        last_step="waiting-step",
        created_by="Fredje",
    )
    success_step = ProcessStepTable(
        process_id=pid, name="generic-step", status="success", state=state, created_by="Fredje"
    )
    waiting_step = ProcessStepTable(
        process_id=pid, name="waiting-step", status="waiting", state="Uberly cool error message", created_by="Fredje"
    )

    db.session.add_all([waiting_workflow, process, success_step, waiting_step])
    db.session.commit()
    return process


@pytest.mark.workflow
def test_resume_workflow(waiting_process):
    with mock.patch("orchestrator.services.processes.resume_process") as m:
        result, process, step_log = run_workflow("task_resume_workflows", {})
        assert_complete(result)
        #
        res = extract_state(result)
        state = {
            "process_id": res["process_id"],
            "reporter": "john.doe",
            "number_of_waiting_processes": 1,
            "number_of_resumed_process_ids": 1,
            "waiting_process_ids": [str(waiting_process.process_id)],
            "resumed_process_ids": [str(waiting_process.process_id)],
        }
        assert_state(result, state)
        m.assert_called_once()


@pytest.mark.workflow
def test_resume_workflow_non_204(waiting_process):
    with mock.patch("orchestrator.services.processes.resume_process") as m:
        m.side_effect = Exception("Failed to resume")

        result, process, step_log = run_workflow("task_resume_workflows", {})
        assert_complete(result)
        #
        res = extract_state(result)
        state = {
            "process_id": res["process_id"],
            "reporter": "john.doe",
            "number_of_waiting_processes": 1,
            "number_of_resumed_process_ids": 0,
            "waiting_process_ids": [str(waiting_process.process_id)],
            "resumed_process_ids": [],
        }
        assert_state(result, state)
        m.assert_called_once()



from orchestrator.graphql.resolvers.customer import resolve_customer
from orchestrator.graphql.resolvers.process import resolve_process, resolve_processes
from orchestrator.graphql.resolvers.product import resolve_products
from orchestrator.graphql.resolvers.product_block import resolve_product_blocks
from orchestrator.graphql.resolvers.resource_type import resolve_resource_types
from orchestrator.graphql.resolvers.settings import SettingsMutation, resolve_settings
from orchestrator.graphql.resolvers.subscription import resolve_subscription, resolve_subscriptions
from orchestrator.graphql.resolvers.workflow import resolve_workflows

__all__ = [
    "resolve_process",
    "resolve_processes",
    "resolve_products",
    "resolve_product_blocks",
    "resolve_settings",
    "SettingsMutation",
    "resolve_subscription",
    "resolve_subscriptions",
    "resolve_customer",
    "resolve_resource_types",
    "resolve_workflows",
]



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def get_non_standard_fields(fields: list[dict]) -> str:
    return ", ".join(type for field in fields if (type := field["type"]) not in ["enum", "int", "bool", "str"])


def is_enum(field: dict) -> bool:
    return field["type"] == "enum"


def is_enum_of_type(field: dict, enum_type: str) -> bool:
    return is_enum(field) and field["enum_type"] == enum_type


def is_str_enum(field: dict) -> bool:
    return is_enum_of_type(field, "str")


def is_int_enum(field: dict) -> bool:
    return is_enum_of_type(field, "int")


def has_default(field: dict) -> bool:
    return "default" in field


def convert_enum(field: dict, separator: str = ".") -> dict:
    return (
        field
        | {"type": (enum_type := field["name"].title().replace("_", ""))}
        | ({"default": enum_type + separator + str(field["default"])} if has_default(field) else {})
    )


def convert_str_enum(field: dict) -> dict:
    return convert_enum(field, separator=".")


def convert_int_enum(field: dict) -> dict:
    return convert_enum(field, separator="._")


def get_str_enums(fields: list[dict]) -> list[dict]:
    return [convert_str_enum(field) for field in fields if is_str_enum(field)]


def get_int_enums(fields: list[dict]) -> list[dict]:
    return [convert_int_enum(field) for field in fields if is_int_enum(field)]


def to_dict(list_of_dicts: list[dict]) -> dict[str, dict]:
    return {dict["name"]: dict for dict in list_of_dicts}



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from uuid import UUID

from sqlalchemy import select
from sqlalchemy.orm import joinedload

from orchestrator.db import ProductTable, db
from orchestrator.types import UUIDstr


def get_products(*, filters: list | None = None) -> list[ProductTable]:
    stmt = select(ProductTable)
    for clause in filters or []:
        stmt = stmt.where(clause)
    return list(db.session.scalars(stmt))


def get_product_by_id(product_id: UUID | UUIDstr, join_fixed_inputs: bool = True) -> ProductTable | None:
    """Get product by id.

    Args:
        product_id: ProductTable id uuid
        join_fixed_inputs: whether to join the fixed inputs in the same query or not

    Returns: ProductTable object

    """
    if not join_fixed_inputs:
        return db.session.get(ProductTable, product_id)

    return db.session.get(ProductTable, product_id, options=[joinedload(ProductTable.fixed_inputs)])


def get_product_by_name(name: str) -> ProductTable:
    stmt = select(ProductTable).options(joinedload(ProductTable.fixed_inputs)).where(ProductTable.name == name)
    return db.session.scalars(stmt).unique().one()


def get_types() -> list[str]:
    stmt = select(ProductTable.product_type).distinct()
    return list(db.session.scalars(stmt))


def get_tags() -> list[str]:
    stmt = select(ProductTable.tag).distinct()
    return list(db.session.scalars(stmt))



import datetime
from copy import deepcopy
from uuid import uuid4

import pytest
from sqlalchemy import select
from sqlalchemy.exc import MultipleResultsFound

from orchestrator.db import ProductTable, db
from orchestrator.domain import SubscriptionModel
from orchestrator.services.subscriptions import (
    build_extended_domain_model,
    format_extended_domain_model,
    get_subscription,
    retrieve_subscription_by_subscription_instance_value,
)
from orchestrator.utils.json import json_dumps, json_loads
from test.unit_tests import fixtures

CORRECT_SUBSCRIPTION = str(uuid4())
INCORRECT_SUBSCRIPTION = str(uuid4())

subscription_mapping = {"PB_2": [{"rt_3": "info.id", "rt_2": "info2.id"}]}


def get_one_product(product_name):
    return db.session.scalars(select(ProductTable).where(ProductTable.name == product_name)).one()


def test_get_subscription_by_id(generic_product_3):
    values = {"info.id": "0", "info2.id": "X"}
    product = get_one_product("Product 3")
    subscription = fixtures.create_subscription_for_mapping(
        product, subscription_mapping, values, subscription_id=CORRECT_SUBSCRIPTION
    )
    result = get_subscription(str(subscription.subscription_id))
    assert str(result.subscription_id) == CORRECT_SUBSCRIPTION


def test_get_subscription_by_id_err(generic_product_3):
    values = {"info.id": "0", "info2.id": "X"}
    product = get_one_product("Product 3")
    fixtures.create_subscription_for_mapping(
        product, subscription_mapping, values, subscription_id=CORRECT_SUBSCRIPTION
    )

    with pytest.raises(ValueError):
        get_subscription(INCORRECT_SUBSCRIPTION)


def test_get_subscription_by_id_invalid_id(generic_product_3):
    values = {"info.id": "0", "info2.id": "X"}
    product = get_one_product("Product 3")
    fixtures.create_subscription_for_mapping(
        product, subscription_mapping, values, subscription_id=CORRECT_SUBSCRIPTION
    )

    with pytest.raises(ValueError):
        get_subscription("abc")


def test_retrieve_subscription_by_subscription_instance_value_none(generic_product_3):
    values = {"info.id": "0", "info2.id": "X"}
    product = get_one_product("Product 3")
    fixtures.create_subscription_for_mapping(product, subscription_mapping, values)

    assert retrieve_subscription_by_subscription_instance_value("rt_2", "Wrong") is None


def test_retrieve_subscription_by_subscription_instance_value(generic_product_3):
    values = {"info.id": "0", "info2.id": "X"}
    product = get_one_product("Product 3")
    fixtures.create_subscription_for_mapping(
        product, subscription_mapping, values, subscription_id=CORRECT_SUBSCRIPTION
    )

    assert (
        str(retrieve_subscription_by_subscription_instance_value("rt_3", "0").subscription_id) == CORRECT_SUBSCRIPTION
    )


def test_retrieve_subscription_by_subscription_instance_value_err(generic_product_3):
    values = {"info.id": "0", "info2.id": "X"}
    product = get_one_product("Product 3")
    fixtures.create_subscription_for_mapping(product, subscription_mapping, values)
    fixtures.create_subscription_for_mapping(product, subscription_mapping, values)

    with pytest.raises(MultipleResultsFound):
        retrieve_subscription_by_subscription_instance_value("rt_3", "0")


def test_build_extended_domain_model(generic_subscription_1, generic_product_1):
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    extended_model = build_extended_domain_model(subscription)
    actual = json_loads(json_dumps(extended_model))

    # Remove dates from the result and just verify their type
    assert isinstance(actual.pop("start_date"), datetime.datetime)
    datetime.datetime.fromisoformat(actual["product"].pop("created_at"))

    assert actual == {
        "customer_descriptions": [],
        "customer_id": "2f47f65a-0911-e511-80d0-005056956c1a",
        "description": "Generic Subscription One",
        "end_date": None,
        "insync": True,
        "note": None,
        "pb_1": {
            "label": None,
            "name": "PB_1",
            "owner_subscription_id": generic_subscription_1,
            "rt_1": "Value1",
            "subscription_instance_id": str(subscription.pb_1.subscription_instance_id),
        },
        "pb_2": {
            "label": None,
            "name": "PB_2",
            "owner_subscription_id": generic_subscription_1,
            "rt_2": 42,
            "rt_3": "Value2",
            "subscription_instance_id": str(subscription.pb_2.subscription_instance_id),
        },
        "product": {
            # "created_at": "2022-12-20T10:36:36+00:00",
            "description": "Generic Product One",
            "end_date": None,
            "name": "Product 1",
            "product_id": str(generic_product_1.product_id),
            "product_type": "Generic",
            "status": "active",
            "tag": "GEN1",
        },
        # "start_date": datetime.datetime(2022, 12, 20, 10, 36, 36, tzinfo=datetime.timezone.utc),
        "status": "active",
        "subscription_id": generic_subscription_1,
    }


def test_format_extended_domain_model(generic_subscription_1, generic_product_1):
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    extended_model = build_extended_domain_model(subscription)

    # For the sake of testing, inject 2 in_use_by_ids on the subscription dict; 1 of which belongs
    # to the owner
    other_instance_id = uuid4()
    in_use_by_ids = [subscription.pb_1.subscription_instance_id, other_instance_id]
    extended_model["pb_1"]["in_use_by_ids"] = in_use_by_ids

    # Without filtering we should see the instance_id from the owner subscription
    formatted_no_filter = format_extended_domain_model(deepcopy(extended_model), filter_owner_relations=False)
    assert sorted(formatted_no_filter["pb_1"]["in_use_by_ids"]) == sorted(in_use_by_ids)

    # With filtering we should not see the instance_id from the owner subscription
    formatted_with_filter = format_extended_domain_model(deepcopy(extended_model), filter_owner_relations=True)
    assert sorted(formatted_with_filter["pb_1"]["in_use_by_ids"]) == [other_instance_id]



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from sqlalchemy import func, select

from orchestrator.db import db
from orchestrator.db.models import ProcessTable
from orchestrator.schedules.scheduling import scheduler
from orchestrator.services.processes import start_process


@scheduler(name="Validate Products and inactive subscriptions", time_unit="day", at="02:30")
def validate_products() -> None:
    uncompleted_products = db.session.scalar(
        select(func.count())
        .select_from(ProcessTable)
        .filter(ProcessTable.workflow.name == "validate_products", ProcessTable.last_status != "completed")
    )
    if not uncompleted_products:
        start_process("task_validate_products")
